// Copyright (c) 2023 ESWIN Authors. All Rights Reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#include <iostream>
#include <cmath>
#include <algorithm>
#include <vector>
#include <math.h>
#include <stdio.h>
#include "half.h"
#include "edla_log.h"
#include "data_compare.h"

static bool isnanf16(uint16_t data)
{
    uint16_t exponent = (data >> 10) & 0x1f;
    uint16_t significand = data & 0x3FF;

    if ((exponent == 0x1f) && (significand != 0)) {
        return true;
    }

    return false;
}

///////////////////////////////////////////////////////////////////////////////
// normalizedEditDistance implementation
bool normalizedEditDistance(const void *data, const void *refData, ES_DATA_PRECISION_E dataType, size_t dataLen,
                            float tolerance)
{
    const uint8_t *outBufData = reinterpret_cast<const uint8_t *>(data);
    const uint8_t *outFileData = reinterpret_cast<const uint8_t *>(refData);

    float of, bf, diff;

    if ((dataType >= ES_PRECISION_INT8) && (dataType <= ES_PRECISION_UINT64)) {
        for (size_t j = 0; j < dataLen; j++) {
            if (outBufData[j] != outFileData[j]) {
                EDLA_LOG_ERR("dst data check error, index: %ld, real data: 0x%x, expect data: 0x%x\n", j, outBufData[j],
                             outFileData[j]);
                return false;
            }
        }
    } else if (dataType == ES_PRECISION_FP16) {
        for (size_t j = 0; j < dataLen; j = j + 2) {
            if (isnanf16(*((uint16_t *)&outBufData[j]))) {
                EDLA_LOG_ERR("data is NAN, index: %ld\n", j);
                return false;
            }
            of = half_float::detail::half2float(*((uint16_t *)&outBufData[j]));
            bf = half_float::detail::half2float(*((uint16_t *)&outFileData[j]));
            diff = std::fabs(of - bf);
            auto tmp = std::max<float>(std::fabs(of), std::fabs(bf));
            float relative_diff = diff / std::max<float>(0.00000001, tmp);
            float min_error = std::min<float>(diff, relative_diff);

            if (min_error > tolerance) {
                EDLA_LOG_ERR("data check error, index: %ld, real data: %f,"
                             "expect data: %f.hex data(0x%x 0x%x) --- (0x%x 0x%x) "
                             "min_error %f diff %f relative_diff %f\n",
                             j, of, bf, outBufData[j], outBufData[j + 1], outFileData[j], outFileData[j + 1], min_error,
                             diff, relative_diff);
                return false;
            }
        }
    } else if (dataType == ES_PRECISION_FP32) {
        int oi = 0, bi = 0;
        for (size_t j = 0; j < dataLen; j = j + 4) {
            of = *((float *)&outBufData[j]);
            bf = *((float *)&outFileData[j]);

            oi = *((int *)&outBufData[j]);
            bi = *((int *)&outFileData[j]);
            diff = std::fabs(of - bf);
            auto tmp = std::max<float>(std::fabs(of), std::fabs(bf));
            float relative_diff = diff / std::max<float>(0.00000001, tmp);
            float min_error = std::min<float>(diff, relative_diff);

            if (isnanf(min_error)) {
                EDLA_LOG_ERR("data is NAN, index: %ld\n", j);
                return false;
            }
            if (min_error > tolerance) {
                EDLA_LOG_ERR("data check error, index: %ld, real data: %f,expect data: %f."
                             " hex data(0x%x 0x%x 0x%x 0x%x) --- (0x%x 0x%x 0x%x 0x%x) 0x%x -- 0x%x\n",
                             j, of, bf, outBufData[j], outBufData[j + 1], outBufData[j + 2], outBufData[j + 3],
                             outFileData[j], outFileData[j + 1], outFileData[j + 2], outFileData[j + 3], oi, bi);
                return false;
            }
        }
    } else {
        EDLA_LOG_ERR("error:dataType:%d\n", dataType);
        return false;
    }

    return true;
}

///////////////////////////////////////////////////////////////////////////////
// cosDistance implementation
template <typename T>
bool cosineSimilarity(const T *vec1, const T *vec2, size_t dataLen, double *result)
{
    double dotProduct = 0;
    double norm1 = 0;
    double norm2 = 0;

    for (size_t i = 0; i < dataLen / sizeof(T); i++) {
        double _a = (double)vec1[i];
        double _b = (double)vec2[i];
        dotProduct += _a * _b;
        norm1 += _a * _a;
        norm2 += _b * _b;
    }

    if (norm1 == 0.0 || norm2 == 0.0) {
        EDLA_LOG_ERR("cosine similarity is not defined whenever one or both "
                     "input vectors are zero-vectors.");
        return false;
    }
    norm1 = sqrt(norm1);
    norm2 = sqrt(norm2);

    result[0] = dotProduct / (norm1 * norm2);
    result[1] = norm1;
    result[2] = norm2;
    return true;
}

static float float16_to_float32(uint16_t value)
{
    uint16_t sign = (value & 0x8000) >> 15;
    uint16_t exponent = (value & 0x7c00) >> 10;
    uint16_t significand = value & 0x03FF;

    union {
        uint32_t u;
        float f;
    } tmp;
    if (exponent == 0) {
        if (significand == 0) {
            // zero
            tmp.u = (sign << 31);
        } else {
            // denormal
            exponent = 0;
            // find non-zero bit
            while ((significand & 0x200) == 0) {
                significand <<= 1;
                exponent++;
            }
            significand <<= 1;
            significand &= 0x3FF;
            tmp.u = (sign << 31) | ((-exponent + (-15 + 127)) << 23) | (significand << 13);
        }
    } else if (exponent == 0x1F) {
        // infinity or NaN
        tmp.u = (sign << 31) | (0xFF << 23) | (significand << 13);
    } else {
        // normalized
        tmp.u = (sign << 31) | ((exponent + (-15 + 127)) << 23) | (significand << 13);
    }

    return tmp.f;
}

bool cosDistance(const void *data, const void *refData, ES_DATA_PRECISION_E dataType, size_t dataLen, float tolerance)
{
    double result[3] = {0};
    std::vector<float> refDataF32;
    std::vector<float> dataF32;
    int32_t dim = 0;
    bool ret = false;

    switch (dataType) {
        case ES_PRECISION_INT8:
            ret = cosineSimilarity(static_cast<const int8_t *>(data), static_cast<const int8_t *>(refData), dataLen,
                                   result);
            break;

        case ES_PRECISION_UINT8:
            ret = cosineSimilarity(static_cast<const uint8_t *>(data), static_cast<const uint8_t *>(refData), dataLen,
                                   result);
            break;

        case ES_PRECISION_INT16:
            ret = cosineSimilarity(static_cast<const int16_t *>(data), static_cast<const int16_t *>(refData), dataLen,
                                   result);
            break;

        case ES_PRECISION_UINT16:
            ret = cosineSimilarity(static_cast<const uint16_t *>(data), static_cast<const uint16_t *>(refData), dataLen,
                                   result);
            break;

        case ES_PRECISION_INT32:
            ret = cosineSimilarity(static_cast<const int32_t *>(data), static_cast<const int32_t *>(refData), dataLen,
                                   result);
            break;

        case ES_PRECISION_UINT32:
            ret = cosineSimilarity(static_cast<const uint32_t *>(data), static_cast<const uint32_t *>(refData), dataLen,
                                   result);
            break;

        case ES_PRECISION_INT64:
            ret = cosineSimilarity(static_cast<const int64_t *>(data), static_cast<const int64_t *>(refData), dataLen,
                                   result);
            break;

        case ES_PRECISION_UINT64:
            ret = cosineSimilarity(static_cast<const uint64_t *>(data), static_cast<const uint64_t *>(refData), dataLen,
                                   result);
            break;
        case ES_PRECISION_FP32:
            ret = cosineSimilarity(static_cast<const float *>(data), static_cast<const float *>(refData), dataLen,
                                   result);
            if (isnanf(result[0]) || isnanf(result[1]) || isnanf(result[2])) {
                EDLA_LOG_ERR("result is NAN:result[0]:%f, result[1]:%f, result[2]:%f\n", result[0], result[1],
                             result[2]);
                return false;
            }
            break;

        case ES_PRECISION_FP16:
            dim = dataLen / sizeof(uint16_t);
            dataF32 = std::vector<float>(dim);
            refDataF32 = std::vector<float>(dim);
            for (int32_t i = 0; i < dim; ++i) {
                if (isnanf16(((uint16_t *)data)[i]) || isnanf16(((uint16_t *)refData)[i])) {
                    EDLA_LOG_ERR("data is NAN, index: %ld\n", i);
                    return false;
                }
                dataF32[i] = float16_to_float32(((uint16_t *)data)[i]);
                refDataF32[i] = float16_to_float32(((uint16_t *)refData)[i]);
            }
            ret = cosineSimilarity(dataF32.data(), refDataF32.data(), dim * sizeof(float), result);
            break;
        default:
            EDLA_LOG_ERR("input data type not support.");
            return false;
    }

    if (!ret) {
        EDLA_LOG_ERR("run cosineSimilarity failed. result[0]:%f, result[1]:%f, result[2]:%f\n", result[0], result[1],
                     result[2]);
        return false;
    }

    if (isinf(result[0]) || isinf(result[1]) || isinf(result[2])) {
        EDLA_LOG_ERR("data is infinity:result[0]:%f, result[1]:%f, result[2]:%f\n", result[0], result[1], result[2]);
        return false;
    }

    if (isnan(result[0]) || isnan(result[1]) || isnan(result[2])) {
        EDLA_LOG_ERR("data is NaN:result[0]:%f, result[1]:%f, result[2]:%f\n", result[0], result[1], result[2]);
        return false;
    }

    double diff = std::fabs(result[1] - result[2]);
    double relativeDiff = diff / result[2];
    EDLA_LOG_INFO("data check:result[0]:%f, result[1]:%f, result[2]:%f, relativeDiff:%f, tolerance:%f, dataType:%d\n",
                  result[0], result[1], result[2], relativeDiff, tolerance, dataType);

    if (result[0] < 1 - tolerance) {
        EDLA_LOG_ERR(
            "data check error:result[0]:%f, result[1]:%f, result[2]:%f, relativeDiff:%f, tolerance:%f, dataType:%d\n",
            result[0], result[1], result[2], relativeDiff, tolerance, dataType);
        return false;
    }
    return true;
}

template <typename T>
void printData(const T *data, const T *refData, size_t dataLen)
{
    for (size_t i = 0; i < dataLen / sizeof(T); i++) {
        EDLA_LOG_INFO("data[%u]:%f refData[%u]:%f\n", i, static_cast<double>(data[i]), i,
                      static_cast<double>(refData[i]));
    }
}

static void printDataFP16(const uint16_t *data, const uint16_t *refData, size_t dataLen)
{
    for (size_t i = 0; i < dataLen / sizeof(uint16_t); i++) {
        EDLA_LOG_INFO("data[%u]:%x refData[%u]:%x\n", i, data[i], i, refData[i]);
    }
}

static void dumpData(const void *data, const void *refData, ES_DATA_PRECISION_E dataType, size_t dataLen)
{
    switch (dataType) {
        case ES_PRECISION_INT8:
            printData(static_cast<const int8_t *>(data), static_cast<const int8_t *>(refData), dataLen);
            break;
        case ES_PRECISION_UINT8:
            printData(static_cast<const uint8_t *>(data), static_cast<const uint8_t *>(refData), dataLen);
            break;
        case ES_PRECISION_INT16:
            printData(static_cast<const int16_t *>(data), static_cast<const int16_t *>(refData), dataLen);
            break;
        case ES_PRECISION_UINT16:
            printDataFP16(static_cast<const uint16_t *>(data), static_cast<const uint16_t *>(refData), dataLen);
            break;
        case ES_PRECISION_INT32:
            printData(static_cast<const int32_t *>(data), static_cast<const int32_t *>(refData), dataLen);
            break;
        case ES_PRECISION_UINT32:
            printData(static_cast<const uint32_t *>(data), static_cast<const uint32_t *>(refData), dataLen);
            break;
        case ES_PRECISION_INT64:
            printData(static_cast<const int64_t *>(data), static_cast<const int64_t *>(refData), dataLen);
            break;
        case ES_PRECISION_UINT64:
            printData(static_cast<const uint64_t *>(data), static_cast<const uint64_t *>(refData), dataLen);
            break;
        case ES_PRECISION_FP32:
            printData(static_cast<const float *>(data), static_cast<const float *>(refData), dataLen);
            break;
        case ES_PRECISION_FP16:
            printDataFP16(static_cast<const uint16_t *>(data), static_cast<const uint16_t *>(refData), dataLen);
            break;
        default:
            EDLA_LOG_ERR("Input data type not supported.");
    }
}

bool dataCompare(const void *data, const void *refData, ES_DATA_PRECISION_E dataType, size_t dataLen, float tolerance)
{
#ifdef DUMP_DATA
    dumpData(data, refData, dataType, dataLen);
#endif
#ifdef COS_DISTANCE
    return cosDistance(data, refData, dataType, dataLen, tolerance);
#else
    return normalizedEditDistance(data, refData, dataType, dataLen, tolerance);
#endif
}
