// Copyright © 2024 ESWIN. All rights reserved.
//
// Beijing ESWIN Computing Technology Co., Ltd and its affiliated companies ("ESWIN") retain
// all intellectual property and proprietary rights in and to this software. Except as expressly
// authorized by ESWIN, no part of the software may be released, copied, distributed, reproduced,
// modified, adapted, translated, or created derivative work of, in whole or in part.

#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>
#include <algorithm>
#include <iterator>
#include <cstring>
#include <filesystem>
#include <getopt.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <pthread.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <math.h>
#include <thread>
#include <chrono>
#include <hwpreprocess.h>
#include "es_ai_inference.h"

static const int32_t MAX_HANDLED_TASK = 50;

template <typename T>
void check(T result, char const *const func, const char *const file, ES_S32 const line)
{
    if (result) {
        printf("error at %s:%d error=%d \"%s\" \n", file, line, static_cast<unsigned ES_S32>(result), func);
        exit(EXIT_FAILURE);
    }
}

#define esCheckError(val) check((val), #val, __FILE__, __LINE__)

using namespace cv;
using namespace std;

void EsAiInference::displayImagesWithLabels(const vector<pair<string, string>> &imagesAndLabels)
{
    const string windowName = "Image Display";
    namedWindow(windowName, WINDOW_AUTOSIZE);

    for (const auto &imageAndLabel : imagesAndLabels) {
        const string &imagePath = imageAndLabel.first;
        const string &label = imageAndLabel.second;

        Mat img = imread(imagePath, IMREAD_COLOR);
        if (img.empty()) {
            cerr << "Could not read the image: " << imagePath << endl;
            continue;
        }

        imshow(windowName, img);
        waitKey(3000);

        int font_face = FONT_HERSHEY_SIMPLEX;
        double font_scale = 0.6;
        int thickness = 2;
        int baseline = 0;

        Size text_size = getTextSize(label, font_face, font_scale, thickness, &baseline);

        Point text_org(img.cols - text_size.width - 10, text_size.height + 10);

        putText(img, label, text_org, font_face, font_scale, Scalar(0, 255, 0), thickness);

        imshow(windowName, img);
        waitKey(3000);
    }

    destroyWindow(windowName);
}

EsAiInference::EsAiInference() : mModelId(0), mTaskCount(0), mStarted(false)
{
    mParams.modelPath = "";
    mParams.inputPath = "";
    mParams.outputPath = "";
    mParams.display = false;

#if (USE_HAE_ACCELERATION == 1)
    mPreprocess = std::make_shared<EsHwPreProcess>();
#else
    mPreprocess = std::make_shared<EsPreProcess>();
#endif
}

EsAiInference::~EsAiInference()
{
    if (!mParams.modelPath.empty() && mStarted) release();
}

void EsAiInference::printHelp()
{
    printf("usage: es_ai_inference --model=string --config=string [options] ...\n");
    printf("options:\n");
    printf("-m, --model                path to a model file (string)\n");
    printf("-c, --config               preprocess configuration (string)\n");
    printf("-C, --classes              classes info file(string), classify network need specify\n");
    printf("-p, --postprocess          postProcess info, default is classify process\n");
    printf("-i, --input-dir            the directory of each inputs (folders) located (string [=])\n");
    printf("-o, --output-dir           the directory of each outputs (folders) will saved in (string [=])\n");
    printf("-l, --list                 the list of inputs which will test (string [=])\n");
    printf("-d  --display-pic          display the image on the screen\n");
    printf("-?, --help                 print this message\n");
}

int32_t EsAiInference::inputParser(int argc, char **argv)
{
    int option;
    struct option longOptions[] = {{"model", required_argument, 0, 'm'},
                                   {"config", required_argument, 0, 'c'},
                                   {"classes", required_argument, 0, 'C'},
                                   {"postprocess", required_argument, 0, 'p'},
                                   {"input-dir", required_argument, 0, 'i'},
                                   {"output-dir", required_argument, 0, 'o'},
                                   {"list", required_argument, 0, 'l'},
                                   {"display-pic", no_argument, 0, 'd'},
                                   {"help", no_argument, 0, '?'},
                                   {0, 0, 0, 0}};

    while ((option = getopt_long(argc, argv, "m:c:C:p:i:o:l:d?", longOptions, nullptr)) != -1) {
        switch (option) {
            case 'm':
                mParams.modelPath = optarg;
                break;
            case 'c':
                mParams.preprocessConfig = optarg;
                break;
            case 'C':
                mParams.classesConfig = optarg;
                break;
            case 'p':
                mParams.postprocessConfig = optarg;
                break;
            case 'i':
                mParams.inputPath = optarg;
                if (mParams.inputPath.back() != '/') {
                    mParams.inputPath += '/';
                }
                break;
            case 'o':
                mParams.outputPath = optarg;
                if (mParams.outputPath.back() != '/') {
                    mParams.outputPath += '/';
                }
                break;
            case 'l':
                mParams.listDir = readListFromFile(optarg);
                break;
            case 'd':
                mParams.display = true;
                break;
            case '?':
            default:
                printHelp();
                return ES_FAILURE;
        }
    }

    if (mParams.modelPath.empty()) {
        printf("Error: Model path (-m) is required.\n");
        printHelp();
        return ES_FAILURE;
    }

    if (mParams.preprocessConfig.empty()) {
        printf("Error: preprocess config path (-c) is required.\n");
        printHelp();
        return ES_FAILURE;
    }
    uint32_t devId = 0;
    if (!mPreprocess->initializeFromJson(devId, mParams.preprocessConfig)) {
        printf("Error: preprocess config parse failed, please check the input config.\n");
        printHelp();
        return ES_FAILURE;
    }

    if (!mParams.postprocessConfig.empty()) {
        if (!mPostprocess.parseArgs(mParams.postprocessConfig)) {
            printf("Error: postprocess config parse failed, please check the input config.\n");
            return ES_FAILURE;
        }
    }

    if (EsPostProcess::CLASSIFY == mPostprocess.getProcessType()) {
        if (mParams.classesConfig.empty()) {
            printf("Error: classes config path (-C) is required.\n");
            printHelp();
            return ES_FAILURE;
        }
        if (!mPostprocess.parseClassification(mParams.classesConfig)) {
            printf("Error: classes config parse failed, please check the input config.\n");
            return ES_FAILURE;
        }
    }

    if (mParams.inputPath.empty()) {
        printf("Error: No input-dir specified;\n");
        printHelp();
        return ES_FAILURE;
    }

    if (mParams.listDir.empty()) {
        std::vector<std::string> binFiles = getJEPGFilesFromListDir(mParams.inputPath);
        if (binFiles.empty()) {
            printf("Error: have no jpeg in %s\n", mParams.inputPath.c_str());
            return ES_FAILURE;
        }
        std::sort(binFiles.begin(), binFiles.end());
        std::filesystem::path inputPath(mParams.inputPath);
        mParams.inputList.push_back({inputPath.stem(), binFiles});
    } else {
        for (const auto &directory : mParams.listDir) {
            std::string fullDirectoryPath = mParams.inputPath + directory;
            std::vector<std::string> binFiles = getJEPGFilesFromListDir(fullDirectoryPath);
            if (binFiles.empty()) {
                printf("Error: have no jpeg in list Dir %s\n", fullDirectoryPath.c_str());
                return ES_FAILURE;
            }
            std::sort(binFiles.begin(), binFiles.end());
            mParams.inputList.push_back({directory, binFiles});
        }
    }

#ifdef DUMP_INPUT_PARAMS
    dumpParams();
#endif

    return ES_SUCCESS;
}

int32_t EsAiInference::startRunModel()
{
    uint16_t deviceNum = 0;
    uint16_t deviceId = 0;

    esCheckError(ES_NPU_GetNumDevices(&deviceNum));
    if (deviceId >= deviceNum) {
        printf("deviceId(%d) is not less than deviceNum(%d)\n", deviceId, deviceNum);
        return ES_NPU_ERROR_DEV_NOT_FOUND;
    }
    esCheckError(ES_NPU_SetDevice(deviceId));
    esCheckError(ES_NPU_LoadModelFromFile(&mModelId, mParams.modelPath.c_str()));
    mStarted = true;

    mTaskList.clear();
    mHandledList.clear();

    auto maxTime = std::chrono::duration<double, std::milli>::min();
    auto minTime = std::chrono::duration<double, std::milli>::max();
    auto totalTime = std::chrono::duration<double, std::milli>::zero();

    do {
        esCheckError(createTaskList());
        if (mTaskList.empty()) {
            printf("taskList is empty\n");
            break;
        }

        for (auto &taskComp : mTaskList) {
            auto start_time = std::chrono::high_resolution_clock::now();
            esCheckError(ES_NPU_Submit(&taskComp->task, 1));
            auto end_time = std::chrono::high_resolution_clock::now();
            auto duration =
                std::chrono::duration_cast<std::chrono::duration<double, std::milli>>(end_time - start_time);
            if (duration > maxTime) maxTime = duration;
            if (duration < minTime) minTime = duration;
            totalTime += duration;
            mTaskCount++;
        }

        esCheckError(inferenceDataProcess());

        for (auto &taskComp : mTaskList) {
            for (int32_t i = 0; i < taskComp->task.inputFdNum; i++) {
                esCheckError(ES_NPU_Unprepare(&taskComp->task.inputFd[i]));
                ES_SYS_MemFree(taskComp->task.inputFd[i].memFd);
            }

            for (int32_t i = 0; i < taskComp->task.outputFdNum; i++) {
                esCheckError(ES_NPU_Unprepare(&taskComp->task.outputFd[i]));
                ES_SYS_Munmap(taskComp->outBufAddr[i], taskComp->outTensor[i].bufferSize);
                ES_SYS_MemFree(taskComp->task.outputFd[i].memFd);
            }
        }

        mTaskList.clear();

        auto newEnd = std::remove_if(mParams.inputList.begin(), mParams.inputList.end(),
                                     [&](const std::pair<std::string, std::vector<std::string>> &item) {
                                         return std::find(mHandledList.begin(), mHandledList.end(), item.first) !=
                                                mHandledList.end();
                                     });

        if (newEnd != mParams.inputList.end()) {
            mParams.inputList.erase(newEnd, mParams.inputList.end());
            mHandledList.clear();
        }
    } while (!mParams.inputList.empty());

    auto averageTime = totalTime / mTaskCount;

    mPerfData.minTime = minTime.count();
    mPerfData.maxTime = maxTime.count();
    mPerfData.avgTime = averageTime.count();

    return ES_SUCCESS;
}

int32_t EsAiInference::handleOutput() { return ES_SUCCESS; }

int32_t EsAiInference::createTask(std::shared_ptr<TaskComponent> taskComp)
{
    int32_t err;
    int32_t numInputTensors;
    int32_t numOutputTensors;
    NPU_TENSOR_S *tensor;
    void *pData;
    ES_U64 memFd;
    static int taskid = 0;

    NPU_TASK_S *task = &taskComp->task;

    esCheckError(ES_NPU_GetNumInputTensors(mModelId, &numInputTensors));
    esCheckError(ES_NPU_GetNumOutputTensors(mModelId, &numOutputTensors));

    task->inputFdNum = numInputTensors;
    task->outputFdNum = numOutputTensors;
    task->modelId = mModelId;
    task->taskId = taskid++;

    for (int bindId = 0; bindId < numInputTensors; bindId++) {
        tensor = &taskComp->inTensor[bindId];

        err = ES_NPU_GetInputTensorDesc(mModelId, bindId, tensor);
        if (ES_SUCCESS != err) {
            printf("getInputTensorDesc error\n");
            return ES_NPU_ERROR_BAD_PARAM;
        }
#if (NPU_DEV_SIM != NPU_REAL_ENV)
        esCheckError(ES_SYS_MemAlloc(&memFd, SYS_CACHE_MODE_NOCACHE, "npu", "reserved", tensor->bufferSize));
#else
        esCheckError(ES_SYS_MemAlloc(&memFd, SYS_CACHE_MODE_NOCACHE, "npu", NULL, tensor->bufferSize));
#endif
        task->inputFd[bindId].memFd = memFd;
        task->inputFd[bindId].size = tensor->bufferSize;
        task->inputFd[bindId].offset = 0;
        // EDLA_LOG_DEBUG("input:bindId=%d memFd=%llx inBufAddr=%p bufferSize=%d dataType=%d\n", bindId, memFd, pData,
        //                tensor->bufferSize, tensor->dataType);
        esCheckError(ES_NPU_Prepare(&task->inputFd[bindId]));
    }

    for (int bindId = 0; bindId < numOutputTensors; bindId++) {
        tensor = &taskComp->outTensor[bindId];

        err = ES_NPU_GetOutputTensorDesc(mModelId, bindId, tensor);
        if (ES_SUCCESS != err) {
            printf("getInputTensorDesc error\n");
            return ES_NPU_ERROR_BAD_PARAM;
        }

#if (NPU_DEV_SIM != NPU_REAL_ENV)
        esCheckError(ES_SYS_MemAlloc(&memFd, SYS_CACHE_MODE_NOCACHE, "npu", "reserved", tensor->bufferSize));
#else
        esCheckError(ES_SYS_MemAlloc(&memFd, SYS_CACHE_MODE_NOCACHE, "npu", NULL, tensor->bufferSize));
#endif

        pData = ES_SYS_Mmap(memFd, tensor->bufferSize, SYS_CACHE_MODE_NOCACHE);
        if (!pData) {
            printf("ES_SYS_Mmap failed!\n");
            return ES_FAILURE;
        }
        memset(pData, 0, tensor->bufferSize);
        task->outputFd[bindId].memFd = memFd;
        task->outputFd[bindId].offset = 0;
        task->outputFd[bindId].size = tensor->bufferSize;
        taskComp->outBufAddr[bindId] = pData;
        // EDLA_LOG_DEBUG("output:bindId=%d memFd=%llx outBufAddr=%p bufferSize=%d dataType=%d\n", bindId, memFd, pData,
        //                tensor->bufferSize, tensor->dataType);
        esCheckError(ES_NPU_Prepare(&task->outputFd[bindId]));
    }

    return ES_SUCCESS;
}

int32_t EsAiInference::createTaskList()
{
    int32_t i = 0;
    size_t inFileLen;
    std::string fullDirectoryPath;
    static uint32_t seed = 256;

    for (const auto &path : mParams.inputList) {
        auto taskComp = std::make_shared<TaskComponent>();
        esCheckError(createTask(taskComp));

        int32_t offset = 0;
        for (const auto &binFiles : path.second) {
            if (mParams.listDir.empty()) {
                fullDirectoryPath = mParams.inputPath + binFiles;
            } else {
                fullDirectoryPath = mParams.inputPath + path.first + "/" + binFiles;
                taskComp->taskDirName = path.first;
            }

            mPreprocess->preprocess(fullDirectoryPath, taskComp->task.inputFd[i].memFd, offset,
                                    taskComp->task.inputFd[i].size / taskComp->inTensor[i].dims.n);

            offset += taskComp->inTensor[i].bufferSize / taskComp->inTensor[i].dims.n;
            if (offset == taskComp->inTensor[i].bufferSize) i++;
        }

        i = 0;
        mTaskList.push_back(taskComp);
        mHandledList.emplace_back(path.first);
        if (mHandledList.size() >= MAX_HANDLED_TASK) {
            break;
        }
    }

    return ES_SUCCESS;
}

std::vector<std::string> EsAiInference::readListFromFile(const std::string &listFilePath)
{
    std::vector<std::string> directories;

    std::ifstream listFile(listFilePath);

    if (!listFile.is_open()) {
        printf("Error: list file %s does not exist or cannot be opened\n", listFilePath.c_str());
        exit(1);
    }

    std::string line;
    while (std::getline(listFile, line)) {
        if (!line.empty()) {
            directories.push_back(line);
        }
    }

    listFile.close();

    return directories;
}

std::vector<std::string> EsAiInference::getBinFilesFromListDir(const std::string &directoryPath)
{
    std::vector<std::string> binFiles;

    if (!std::filesystem::is_directory(directoryPath)) {
        printf("Error: %s in list Dir is not directory\n", directoryPath.c_str());
        exit(1);
    }

    std::filesystem::path dirPath(directoryPath);

    for (const auto &entry : std::filesystem::directory_iterator(dirPath)) {
        if (entry.is_regular_file() && entry.path().extension() == ".bin") {
            binFiles.push_back(entry.path().filename());
        }
    }

    return binFiles;
}

std::vector<std::string> EsAiInference::getJEPGFilesFromListDir(const std::string &directoryPath)
{
    std::vector<std::string> jpegFiles;

    if (!std::filesystem::is_directory(directoryPath)) {
        printf("Error: %s in list Dir is not directory\n", directoryPath.c_str());
        exit(1);
    }

    std::filesystem::path dirPath(directoryPath);

    for (const auto &entry : std::filesystem::directory_iterator(dirPath)) {
        if (entry.is_regular_file() && entry.path().extension() == ".jpeg" ||
            entry.is_regular_file() && entry.path().extension() == ".JPEG" ||
            entry.is_regular_file() && entry.path().extension() == ".jpg") {
            jpegFiles.push_back(entry.path().filename());
        }
    }

    return jpegFiles;
}

void EsAiInference::dumpParams()
{
    printf("listDir: ");
    for (const auto &dir : mParams.listDir) {
        printf("%s ", dir.c_str());
    }
    printf("\n");

    printf("modelPath: %s\n", mParams.modelPath.c_str());

    printf("inputList: \n");
    for (const auto &entry : mParams.inputList) {
        printf("%s: ", entry.first.c_str());
        for (const auto &file : entry.second) {
            printf("%s ", file.c_str());
        }
        printf("\n");
    }

    printf("inputPath: %s\n", mParams.inputPath.c_str());
    printf("outputPath: %s\n", mParams.outputPath.c_str());
}

ES_DATA_PRECISION_E EsAiInference::convertDataType(uint8_t dataType)
{
    switch (dataType) {
        case ES_NPU_DATA_TYPE_FLOAT:
            return ES_PRECISION_FP32;
        case ES_NPU_DATA_TYPE_HALF:
            return ES_PRECISION_FP16;
        case ES_NPU_DATA_TYPE_INT16:
            return ES_PRECISION_INT16;
        case ES_NPU_DATA_TYPE_INT8:
            return ES_PRECISION_INT8;
        case ES_NPU_DATA_TYPE_UINT8:
            return ES_PRECISION_UINT8;
        case ES_NPU_DATA_TYPE_UINT16:
            return ES_PRECISION_UINT16;
        case ES_NPU_DATA_TYPE_INT32:
            return ES_PRECISION_INT32;
        case ES_NPU_DATA_TYPE_UINT32:
            return ES_PRECISION_UINT32;
        case ES_NPU_DATA_TYPE_INT64:
            return ES_PRECISION_INT64;
        case ES_NPU_DATA_TYPE_UINT64:
            return ES_PRECISION_UINT64;
        case ES_NPU_DATA_TYPE_UNKNOWN:
        default:
            printf("data type:%d not support\n", dataType);
            return ES_PRECISION_UNKNOWN;
    }
}

int32_t EsAiInference::inferenceDataProcess()
{
    size_t outFileLen;
    vector<pair<string, string>> imagesAndLabels;

    for (auto &taskComp : mTaskList) {
        std::vector<ES_TENSOR_S> inTensor;
        std::vector<std::vector<int>> topKIndex;
        std::vector<std::vector<float>> topKConfidence;

        for (unsigned long i = 0; i < taskComp->task.outputFdNum; i++) {
            ES_TENSOR_S tensor;
            tensor.pData = taskComp->task.outputFd[i];
            tensor.dataType = mPostprocess.convertDataType(taskComp->outTensor[i].dataType);
            tensor.shapeDim = 6;
            tensor.shape[0] = taskComp->outTensor[i].dims.n;
            tensor.shape[1] = taskComp->outTensor[i].dims.c;
            tensor.shape[2] = taskComp->outTensor[i].dims.h;
            tensor.shape[3] = taskComp->outTensor[i].dims.w;
            tensor.shape[4] = 1;
            tensor.shape[5] = tensor.shape[4] * tensor.shape[1];
            inTensor.push_back(tensor);
        }

        int32_t processType = mPostprocess.getProcessType();
        if (processType == EsPostProcess::CLASSIFY) {
            if (mPostprocess.classify(inTensor, topKIndex, topKConfidence) != 0) {
                printf("classify failed\n");
                return -1;
            }
            std::string searchKey = taskComp->taskDirName;
            auto it = std::find_if(mParams.inputList.begin(), mParams.inputList.end(),
                                   [&searchKey](const std::pair<std::string, std::vector<std::string>> &pair) {
                                       return pair.first == searchKey;
                                   });
            int32_t numInput = taskComp->outTensor[0].dims.n;
            if (it != mParams.inputList.end()) {
                numInput = it->second.size();
            }

            for (int i = 0; i < numInput; i++) {
                std::string label = mPostprocess.getLabelByIdx(topKIndex[i][0]);
                if (!mParams.display)
                    printf("%-15s -> class: %-4d (%-10s), confidence: %-6.2f\n", it->second[i].c_str(), topKIndex[i][0],
                           label.c_str(), topKConfidence[i][0]);

                label = std::to_string(topKIndex[i][0]) + ". " + label;
                label += " confidence: " + std::to_string(topKConfidence[i][0]);
                std::string imgPath = mParams.inputPath + taskComp->taskDirName + "/" + it->second[i];
                imagesAndLabels.emplace_back(imgPath, label);
            }

        } else if (processType == EsPostProcess::DETECTION_OUT) {
            // TODO: Add detection out
        } else if (processType == EsPostProcess::SEGMENT) {
            // TODO: Add segment
        }
    }

    if (mParams.display) displayImagesWithLabels(imagesAndLabels);

    return ES_SUCCESS;
}

int32_t EsAiInference::release()
{
    uint16_t deviceId = 0;

    mPreprocess->release();
    esCheckError(ES_NPU_UnloadModel(mModelId));
    esCheckError(ES_NPU_ReleaseDevice(deviceId));

    return ES_SUCCESS;
}
