// Copyright © 2024 ESWIN. All rights reserved.
//
// Beijing ESWIN Computing Technology Co., Ltd and its affiliated companies ("ESWIN") retain
// all intellectual property and proprietary rights in and to this software. Except as expressly
// authorized by ESWIN, no part of the software may be released, copied, distributed, reproduced,
// modified, adapted, translated, or created derivative work of, in whole or in part.

#ifndef ES_RUN_MODEL_H
#define ES_RUN_MODEL_H

#include <list>
#include <string>
#include <vector>
#include <unordered_map>
#include <memory>
#include <utility>
#include "es_npu_interface.h"
#include "es_sys_memory.h"
#include "preprocess.h"
#include "postprocess.h"
#include <opencv2/opencv.hpp>
#include <map>

typedef struct {
    NPU_TASK_S task;
    void *inBufAddr[ES_TASK_MAX_FD_CNT];
    void *outBufAddr[ES_TASK_MAX_FD_CNT];
    NPU_TENSOR_S inTensor[ES_TASK_MAX_FD_CNT];
    NPU_TENSOR_S outTensor[ES_TASK_MAX_FD_CNT];
    std::string taskDirName;
} TaskComponent;

typedef struct {
    double maxTime;
    double minTime;
    double avgTime;
} PerfData;

using TASK_LIST = std::list<std::shared_ptr<TaskComponent>>;
using FILE_LIST = std::vector<std::pair<std::string, std::vector<std::string>>>;

typedef struct {
    bool display;
    std::vector<std::string> listDir;
    std::string modelPath;
    std::string preprocessConfig;
    std::string postprocessConfig;
    std::string classesConfig;
    std::string inputPath;
    std::string outputPath;
    FILE_LIST inputList;
} InferenceParameters;

class EsAiInference
{
 public:
    EsAiInference();
    virtual ~EsAiInference();

    int32_t inputParser(int argc, char **argv);
    int32_t startRunModel();
    int32_t handleOutput();

 private:
    int32_t inferenceDataProcess();
    int32_t createTaskList();

    std::vector<std::string> readListFromFile(const std::string &listFilePath);
    std::vector<std::string> getBinFilesFromListDir(const std::string &directoryPath);
    std::vector<std::string> getJEPGFilesFromListDir(const std::string &directoryPath);
    int32_t createTask(std::shared_ptr<TaskComponent> taskComp);

    ES_DATA_PRECISION_E convertDataType(uint8_t dataType);
    int32_t release();

    void dumpParams();
    void printHelp();

    void displayImagesWithLabels(const std::vector<std::pair<std::string, std::string>>& imagesAndLabels);

 private:
    uint32_t mModelId;
    TASK_LIST mTaskList;
    PerfData mPerfData;
    uint32_t mTaskCount;
    InferenceParameters mParams;
    std::vector<std::string> mHandledList;
    std::shared_ptr<EsPreProcess> mPreprocess;
    EsPostProcess mPostprocess;
    std::map<std::string, cv::Mat> mMatMap;
    bool mStarted;
};

#endif  // ES_RUN_MODEL_H