// Copyright (c) 2023 ESWIN Authors. All Rights Reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#ifndef ESWIN_HOST_RUNTIME_INCLUDE_CORE_H_
#define ESWIN_HOST_RUNTIME_INCLUDE_CORE_H_

#include <string>
#include <vector>
#include <queue>
#include <unordered_map>
#include <memory>
#include <map>
#include <cstring>
#include <atomic>
#include <condition_variable>
#include <mutex>
#include "list.h"
#include "es_npu_types.h"
#include "esTimer.h"

namespace eswin
{
#define MAGIC_CONTEXT 0x63746578  // "ctex"
#define MAGIC_STREAM 0x7374726d   // "strm"

enum streamType {
    STREAM_TYPE_SYNC = 0,
    STREAM_TYPE_ASYNC_BLOCK,
    STREAM_TYPE_ASYNC_NONBLOCK,
};

enum streamState {
    STREAM_STATE_NORMAL = 0,
    STREAM_STATE_SYNC,
    STREAM_STATE_ABORT,
    STREAM_STATE_DESTROY,
};

typedef struct _npuContext npuContext_t;

typedef struct _npuStream {
    int32_t magic;
    npuContext_t *context;
    enum streamType type;  // SYNC or ASYNC_BLOCK or ASYNC_NONBLOCK
    struct list_head taskList;
    std::atomic_uint32_t taskSubmittedCnt;
    std::atomic_uint32_t taskCompletedCnt;
    std::atomic<streamState> state;
    std::mutex mutex;
    std::condition_variable cond;
    std::mutex syncMutex;
    std::condition_variable syncCond;
} npuStream_t;

class TaskScheduler;
class HWProcessorMgr;
typedef struct _npuContext {
    int32_t magic;
    int32_t deviceId;                      // dieId
    class HWProcessorMgr *hwProcessorMgr;  // devFd资源统一由HWProcessorMgr管理
    class TaskScheduler *taskScheduler;
    std::vector<npuStream_t *> vecStream;
    struct list_head taskList; /*存放所有在该context中提交任务的链表头*/
    std::mutex mutex;
    std::condition_variable cond;
} npuContext_t;

int32_t createContext(npu_context *context, uint16_t deviceId);
int32_t setCurrentContext(npu_context context);
int32_t getCurrentContext(npu_context *context);
int32_t getCurrentContextInner(npu_context *context);
int32_t destroyContextInner(npuContext_t *ctx);
int32_t destroyContext(npu_context context);
int32_t createStream(npu_stream *stream);
int32_t createStreamWithCtx(npuStream_t **stream, npuContext_t *ctx);
int32_t destroyStream(npu_stream stream);
int32_t synchronizeStream(npu_stream stream);
int32_t abortStream(npu_stream stream);
int32_t submitSync(NPU_TASK_S *tasks, uint32_t numTasks);
int32_t submitAsync(NPU_TASK_S *tasks, uint32_t numTasks, npu_stream stream);
int32_t setFlexibleTaskAttribute(uint32_t modelId, NPU_FLEXIBLE_TASK_ATTR_S *taskAttr);
int32_t allocTaskMemory(uint32_t modelId, uint32_t nums, NPU_TASK_MEM_S *taskMem);
int32_t releaseTaskMemory(uint32_t modelId, uint32_t nums, NPU_TASK_MEM_S *taskMem);
int32_t submitFlexibleTask(NPU_TASK_S *tasksIn, ES_U32 numTasks, npu_stream stream);
int32_t processReport(npu_stream stream, int32_t timeoutMs);
int32_t initScheduler(npuContext_t *ctx);
int32_t uninitScheduler(npuContext_t *ctx);
int32_t prepare(ES_DEV_BUF_S *devBuf);
int32_t unprepare(ES_DEV_BUF_S *devBuf);

}  // namespace eswin

#endif  // ESWIN_HOST_RUNTIME_INCLUDE_CORE_H_
