// Copyright (c) 2023 ESWIN Authors. All Rights Reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#ifndef ESWIN_HOST_RUNTIME_INCLUDE_ILOADABLE_H_
#define ESWIN_HOST_RUNTIME_INCLUDE_ILOADABLE_H_

#include <string>
#include <vector>
#include <cstring>
#include "loadableType.h"
#include "iLoadableEntry.h"
#include "iType.h"
#include "edla_log.h"

// some gnu stuff is defining these in host mode... evil.
#ifdef major
#undef major
#endif
#ifdef minor
#undef minor
#endif

namespace eswin
{

class ILoadable
{
 public:
    enum Interface {
        Interface_NONE = ESWIN_LOADABLE_INTERFACE_NONE,
        Interface_UMD = ESWIN_LOADABLE_INTERFACE_UMD,
        Interface_KMD = ESWIN_LOADABLE_INTERFACE_KMD,
    };

    enum MemoryDomain {
        MemoryDomain_SYSMEM = ESWIN_LOADABLE_MEMORY_DOMAIN_SYSMEM,
        MemoryDomain_SRAM = ESWIN_LOADABLE_MEMORY_DOMAIN_SRAM,
    };

    enum MemoryFlags {
        MemoryFlags_NONE = ESWIN_LOADABLE_MEMORY_FLAGS_NONE,
        MemoryFlags_ALLOC = ESWIN_LOADABLE_MEMORY_FLAGS_ALLOC,
        MemoryFlags_SET = ESWIN_LOADABLE_MEMORY_FLAGS_SET,
        MemoryFlags_INPUT = ESWIN_LOADABLE_MEMORY_FLAGS_INPUT,
        MemoryFlags_OUTPUT = ESWIN_LOADABLE_MEMORY_FLAGS_OUTPUT,
        MemoryFlags_DEBUG = ESWIN_LOADABLE_MEMORY_FLAGS_DEBUG,
        MemoryFlags_CACHE = ESWIN_LOADABLE_MEMORY_FLAGS_CACHE,
        MemoryFlags_LLC = ESWIN_LOADABLE_MEMORY_FLAGS_LLC
    };

    enum EventOp {
        EventOp_WAIT = ESWIN_LOADABLE_EVENT_OP_WAIT,
        EventOp_SIGNAL = ESWIN_LOADABLE_EVENT_OP_SIGNAL
    };

    struct Version {
        uint8_t major;
        uint8_t minor;
        uint8_t sub_minor;
        Version(uint8_t maj, uint8_t min, uint8_t sub) : major(maj), minor(min), sub_minor(sub) {}
        Version() : major(0), minor(0), sub_minor(0) {}

        void toC(esLoadableVersion &c) const
        {
            c.major = major;
            c.minor = minor;
            c.subMinor = sub_minor;
        }
    };

    struct MemoryListEntry {
        uint16_t id;
        uint64_t size;
        uint32_t alignment;  // 0 for n/a, otherwise byte alignment
        uint8_t domain;
        static inline uint8_t domain_sysmem() { return MemoryDomain_SYSMEM; }
        static inline uint8_t domain_sram() { return MemoryDomain_SRAM; }
        uint8_t flags;  // alloc or alloc_content or is-input or is-output
        static inline uint8_t flags_alloc() { return MemoryFlags_ALLOC; }
        static inline uint8_t flags_set() { return MemoryFlags_SET; }
        static inline uint8_t flags_input() { return MemoryFlags_INPUT; }
        static inline uint8_t flags_output() { return MemoryFlags_OUTPUT; }
        static inline uint8_t flags_debug() { return MemoryFlags_DEBUG; }
        static inline uint8_t flags_cache() { return MemoryFlags_CACHE; }
        static inline uint8_t flags_llc() { return MemoryFlags_LLC; }
        uint16_t bind_id;                   // valid if flag_{input|output|debug}()  is set
        uint16_t tensor_desc_id;            // valid if bind_id is valid ( != -1 )
        std::vector<std::string> contents;  // symbolic reference to content blob
        std::vector<uint64_t> offsets;      // associated offset for contents

        MemoryListEntry()
            : id(0), size(0), alignment(0), domain(0), flags(0), bind_id(0), tensor_desc_id(0), contents(), offsets()
        {
        }
        MemoryListEntry(const MemoryListEntry &o)
            : id(o.id),
              size(o.size),
              alignment(o.alignment),
              domain(o.domain),
              flags(o.flags),
              bind_id(o.bind_id),
              tensor_desc_id(o.tensor_desc_id),
              contents(o.contents),
              offsets(o.offsets)
        {
        }
        MemoryListEntry(uint16_t i, uint64_t s, uint32_t a, uint8_t d, uint8_t f, std::string sym = std::string(),
                        uint64_t o = 0)
            : id(i), size(s), alignment(a), domain(d), flags(f), bind_id(0), tensor_desc_id(0)
        {
            if (sym.size()) {
                contents.push_back(sym);
                offsets.push_back(o);
            }
        }
    };

    struct EventListEntry {
        uint16_t id;
        uint16_t target;
        uint8_t op;
        static inline uint8_t op_wait() { return EventOp_WAIT; }
        static inline uint8_t op_signal() { return EventOp_SIGNAL; }
        uint32_t val;
        void toC(esLoadableEventListEntry &c) const
        {
            c.id = id;
            c.target = target;
            c.op = op;
            c.val = val;
        }
    };

    struct SubModelListEntry {
        uint16_t id;
        int16_t kmd_id;
        uint32_t interface;  // DLA interface id
        static inline uint32_t interface_NONE() { return Interface_NONE; }
        static inline uint32_t interface_UMD() { return Interface_UMD; }
        static inline uint32_t interface_KMD() { return Interface_KMD; }

        static inline int16_t kmd_id_NONE() { return -1; }

        std::vector<uint16_t> address_list;  // [addr list id]...[addr list id]
        SubModelListEntry(const SubModelListEntry &o)
            : id(o.id), interface(o.interface), kmd_id(o.kmd_id), address_list(o.address_list)
        {
        }

        SubModelListEntry() : id(0), interface(Interface_NONE), kmd_id(-1), address_list() {}
    };

    struct SubmitListEntry {
        uint16_t id;
        std::vector<uint16_t> tasks;
    };

    struct AddressListEntry {
        uint16_t id;      // all possible address list entries are given an id
        uint16_t mem_id;  // determines hRm (+offset from below)
        uint64_t size;    // assert size <= memory[mem_id].size
        uint64_t offset;  // assert (offset + size) <= memory[mem_id].size
        AddressListEntry() : id(0), mem_id(0), size(0), offset(0) {}
        AddressListEntry(uint16_t i, uint16_t m, uint64_t s, uint64_t o = 0) : id(i), mem_id(m), size(s), offset(o) {}
        AddressListEntry(const AddressListEntry &o) : id(o.id), mem_id(o.mem_id), size(o.size), offset(o.offset) {}
        void toC(esLoadableAddressListEntry &c) const
        {
            c.id = id;
            c.memId = mem_id;
            c.size = size;
            c.offset = offset;
        }
    };

    struct TensorDescListEntry {
        std::string name;
        uint16_t id;
        uint16_t memId;
        uint64_t size;
        uint64_t offset;
        NPU_DIMS4_S dims;
        uint8_t dataFormat;
        uint8_t dataType;
        uint8_t dataCategory;
        uint8_t pixelFormat;
        uint8_t pixelMapping;
        uint32_t stride[ESWIN_LOADABLE_TENSOR_DESC_NUM_STRIDES];
    };

    struct RelocEntry {
        uint16_t addressListId;  // fix vs. this addr list item
        uint16_t writeId;        // fix *within this* memory id given offset below
        uint64_t offset;         // buffer offset to the fixup
        uint32_t interface;      // dla1, emu1, etc.
        uint32_t subInterface;   //  dla1-surf_desc, etc.
        uint8_t relocType;       // stride0..7 (aka line, surf)

        RelocEntry(const RelocEntry &o)
            : addressListId(o.addressListId),
              writeId(o.writeId),
              offset(o.offset),
              interface(o.interface),
              subInterface(o.subInterface),
              relocType(o.relocType)
        {
        }

        RelocEntry(int16_t a, uint64_t o, uint32_t i, uint32_t s, uint8_t r)
            : addressListId(a),
              writeId(0),  // invalid
              offset(o),
              interface(i),
              subInterface(s),
              relocType(r)
        {
        }

        RelocEntry(int16_t a, int16_t w, uint64_t o, uint32_t i, uint32_t s, uint8_t r)
            : addressListId(a), writeId(w), offset(o), interface(i), subInterface(s), relocType(r)
        {
        }
    };

    struct Blob {
        std::string name;
        uint64_t size;
        Interface interface;
        uint32_t subInterface;
        Version version;

        Blob() : size(0), interface(Interface_NONE), subInterface(0) {}

        Blob(const std::string &n, uint64_t s, Interface i, uint32_t si, Version v)
            : name(n), size(s), interface(i), subInterface(si), version(v)
        {
        }
    };

    virtual std::string getName() const = 0;

    virtual int getNumMemoryListEntries() const = 0;
    virtual MemoryListEntry getMemoryListEntry(uint16_t mem_id) const = 0;

    virtual int getNumEventListEntries() const = 0;
    virtual EventListEntry getEventListEntry(uint16_t event_id) const = 0;

    virtual int getNumSubModelListEntries() const = 0;
    virtual SubModelListEntry getSubModelListEntry(uint16_t task_id) const = 0;

    virtual int getNumAddressListEntries() const = 0;
    virtual AddressListEntry getAddressListEntry(uint16_t i) const = 0;

    virtual int getNumTensorDescListEntries() const = 0;
    virtual TensorDescListEntry getTensorDescListEntry(uint16_t i) const = 0;

    virtual int32_t getNetworkDataType(DataType::UnderlyingType *) const = 0;

    virtual int32_t getNumInputTensors(int *) const = 0;
    virtual int32_t getInputTensorDesc(uint16_t id, ILoadable::TensorDescListEntry *) const = 0;

    virtual int32_t getNumOutputTensors(int *) const = 0;
    virtual int32_t getOutputTensorDesc(uint16_t id, ILoadable::TensorDescListEntry *) const = 0;

 protected:
    ILoadable();
    virtual ~ILoadable();
};

}  // namespace eswin

#endif  // ESWIN_HOST_RUNTIME_INCLUDE_ILOADABLE_H_
