// Copyright (c) 2023 ESWIN Authors. All Rights Reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#ifndef ESWIN_HOST_RUNTIME_INCLUDE_ITYPE_H_
#define ESWIN_HOST_RUNTIME_INCLUDE_ITYPE_H_

#include <stdexcept>
#include <string>
#include <map>

#include "es_npu_error.h"

#include "loadableType.h"

#define ENUM_CLASS_MEMBERS(ENUM_C)                                                \
 public:                                                                          \
    ENUM_C(UnderlyingType v) : m_v(v)                                             \
    {                                                                             \
        if (m_v > max()) throw std::out_of_range(#ENUM_C);                        \
    }                                                                             \
    ENUM_C(const ENUM_C& o) : m_v(o.m_v) {}                                       \
    ENUM_C(Enum e) : m_v(UnderlyingType(e)) {}                                    \
    ENUM_C() : m_v(UnderlyingType(0)) {} /*note default is zero*/                 \
    bool operator==(const ENUM_C& rhs) const { return m_v == rhs.m_v; }           \
    bool operator==(const Enum& rhs) const { return m_v == UnderlyingType(rhs); } \
    operator UnderlyingType() const { return m_v; }                               \
    UnderlyingType v() const { return m_v; }                                      \
                                                                                  \
 protected:                                                                       \
    UnderlyingType m_v;

namespace eswin
{

template <typename T>
inline int EnumMax();  // used by checkers

class DataFormat
{
 public:
    typedef uint8_t UnderlyingType;
    enum Enum {
        UNKNOWN = ESWIN_DATA_FORMAT_UNKNOWN,
        NCHW = ESWIN_DATA_FORMAT_NCHW,
        NHWC = ESWIN_DATA_FORMAT_NHWC,
        NCxHWx = ESWIN_DATA_FORMAT_NCxHWx,
    };
    static inline UnderlyingType max() { return 3U; }
    const char* c_str() const
    {
        const char* names[4] = {"UNKNOWN", "NCHW", "NHWC", "NCxHWx"};
        return names[m_v];
    }

    ENUM_CLASS_MEMBERS(DataFormat);
};
template <>
inline int EnumMax<DataFormat>()
{
    return DataFormat::max() + 1;
}

class DataType
{
 public:
    typedef uint8_t UnderlyingType;
    enum Enum {
        UNKNOWN = ESWIN_DATA_TYPE_UNKNOWN,
        FLOAT = ESWIN_DATA_TYPE_FLOAT,    //!< FP32 format
        HALF = ESWIN_DATA_TYPE_HALF,      //!< FP16 format
        INT16 = ESWIN_DATA_TYPE_INT16,    //!< INT16 format
        INT8 = ESWIN_DATA_TYPE_INT8,      //!< INT8 format
        UINT8 = ESWIN_DATA_TYPE_UINT8,    //!< UINT8 format
        UINT16 = ESWIN_DATA_TYPE_UINT16,  //!< UINT16 format
    };
    static inline UnderlyingType max() { return 6U; }
    const char* c_str() const
    {
        const char* names[7] = {"UNKNOWN", "FLOAT", "HALF", "INT16", "INT8", "UINT8", "UINT16"};
        return names[m_v];
    }
    static UnderlyingType getEnum(std::string find)
    {
        uint8_t i = 0;
        const char* names[7] = {"UNKNOWN", "FLOAT", "HALF", "INT16", "INT8", "UINT8", "UINT16"};
        uint8_t match = UNKNOWN;
        for (; i <= max(); ++i) {
            if (names[i] == find) {
                match = i;
                break;
            }
        }
        return match;
    }
    ENUM_CLASS_MEMBERS(DataType);
};
template <>
inline int EnumMax<DataType>()
{
    return DataType::max() + 1;
}  // used by checkers

class TensorScalingMode
{
 public:
    typedef uint8_t UnderlyingType;
    enum Enum {
        NONE = ESWIN_TENSOR_SCALING_MODE_NONE,
        PER_TENSOR = ESWIN_TENSOR_SCALING_MODE_PER_TENSOR,
        PER_CHANNEL = ESWIN_TENSOR_SCALING_MODE_PER_CHANNEL,
    };
    static inline UnderlyingType max() { return 2U; }
    const char* c_str() const
    {
        const char* names[3] = {"NONE", "PER_TENSOR", "PER_CHANNEL"};
        return names[m_v];
    }
    static UnderlyingType getEnum(std::string find)
    {
        uint8_t i = 0;
        const char* names[3] = {"NONE", "PER_TENSOR", "PER_CHANNEL"};
        uint8_t match = NONE;
        for (; i <= max(); ++i) {
            if (names[i] == find) {
                match = i;
                break;
            }
        }
        return match;
    }
    ENUM_CLASS_MEMBERS(TensorScalingMode);
};
template <>
inline int EnumMax<TensorScalingMode>()
{
    return TensorScalingMode::max() + 1;
}  // used by checkers

class QuantizationMode
{
 public:
    typedef uint8_t UnderlyingType;
    enum Enum {
        NONE = ESWIN_TENSOR_QUANTIZATION_MODE_NONE,
        PER_KERNEL = ESWIN_TENSOR_QUANTIZATION_MODE_PER_KERNEL,  // per KCRS
        PER_FILTER = ESWIN_TENSOR_QUANTIZATION_MODE_PER_FILTER,  // per CRS
    };
    static inline UnderlyingType max() { return 2U; }
    const char* c_str() const
    {
        const char* names[3] = {"NONE", "PER_KERNEL", "PER_FILTER"};
        return names[m_v];
    }
    static UnderlyingType getEnum(std::string find)
    {
        uint8_t i = 0;
        const char* names[3] = {"NONE", "PER_KERNEL", "PER_FILTER"};
        uint8_t match = NONE;
        for (; i <= max(); ++i) {
            if (names[i] == find) {
                match = i;
                break;
            }
        }
        return match;
    }
    ENUM_CLASS_MEMBERS(QuantizationMode);
};
template <>
inline int EnumMax<QuantizationMode>()
{
    return QuantizationMode::max() + 1;
}  // used by checkers

class DataCategory
{
 public:
    typedef uint8_t UnderlyingType;
    enum Enum {
        IMAGE = ESWIN_DATA_CATEGORY_IMAGE,
        WEIGHT = ESWIN_DATA_CATEGORY_WEIGHT,
        FEATURE = ESWIN_DATA_CATEGORY_FEATURE,
        PLANAR = ESWIN_DATA_CATEGORY_PLANAR,
        BIAS = ESWIN_DATA_CATEGORY_BIAS,
    };
    static inline UnderlyingType max() { return 4U; }
    const char* c_str() const
    {
        const char* names[5] = {"IMAGE", "WEIGHT", "FEATURE", "PLANAR", "BIAS"};
        return names[m_v];
    }
    ENUM_CLASS_MEMBERS(DataCategory);
};
template <>
inline int EnumMax<DataCategory>()
{
    return DataCategory::max() + 1;
}  // used by checkers

//
// the type of surface a raw tensor represents
// defined in eswin::priv namespace
// strictly used in pre-dla context
//
enum TensorType {
    kUNKNOWN = 0,
    kNW_INPUT = 1,   /* network input tensor */
    kNW_OUTPUT = 2,  /* network output tensor */
    kIO = 3,         /* input-output tensor passed between operations */
    kWEIGHT = 4,     /* conv weights */
    kBIAS = 5,       /* bias data */
    kBATCH_NORM = 6, /* mean/variance data */
    kSCALE = 7,      /* scaling data */
    kSTREAM = 8,     /* tensors on wire */
    kDEBUG = 9,      /* debug output tensors */
};
template <>
inline int EnumMax<TensorType>()
{
    return 10;
}

class Dims2
{
 public:
    Dims2() : h(0), w(0) {}
    Dims2(int32_t h, int32_t w) : h(h), w(w) {}
    int32_t h;
    int32_t w;
    inline bool operator==(const Dims2& other) const { return (h == other.h && w == other.w); }
    inline bool operator!=(const Dims2& other) const { return !(h == other.h && w == other.w); }
};

class Dims3
{
 public:
    Dims3() : c(0), h(0), w(0) {}
    Dims3(int32_t c, int32_t h, int32_t w) : c(c), h(h), w(w) {}
    int32_t c;  // channels
    int32_t h;
    int32_t w;
};

class Dims4
{
 public:
    Dims4() : n(1), c(0), h(0), w(0) {}
    Dims4(int32_t c, int32_t h, int32_t w) : n(1), c(c), h(h), w(w) {}
    Dims4(int32_t n, int32_t c, int32_t h, int32_t w) : n(n), c(c), h(h), w(w) {}
    int32_t n;  // !< the number of images in the data or number of kernels in the weights (default = 1)
    int32_t c;  // !< the number of channels in the data
    int32_t h;  // !< the height of the data
    int32_t w;  // !< the width of the data
    inline bool operator==(const Dims4& other) const
    {
        return (n == other.n && c == other.c && h == other.h && w == other.w);
    }
    inline bool operator!=(const Dims4& other) const
    {
        return !(n == other.n && c == other.c && h == other.h && w == other.w);
    }
};

class Weights
{
 public:
    Weights() : type(DataType::UNKNOWN), values(NULL), count(0) {}
    Weights(DataType type, const void* values, int64_t count) : type(type), values(values), count(count) {}
    Weights(const Weights& other) : type(other.type), values(other.values), count(other.count) {}

    DataType type;       // !< the type of the weights
    const void* values;  // !< the weight values, in a contiguous array
    int64_t count;       // !< the number of weights in the array
};

class LayerType
{
 public:
    typedef uint8_t UnderlyingType;
    enum Enum {
        kCONVOLUTION = ESWIN_LAYER_TYPE_CONVOLUTION,          //!< Convolution layer
        kFULLY_CONNECTED = ESWIN_LAYER_TYPE_FULLY_CONNECTED,  //!< Fully connected layer
        kACTIVATION = ESWIN_LAYER_TYPE_ACTIVATION,            //!< Activation layer
        kPOOLING = ESWIN_LAYER_TYPE_POOLING,                  //!< Pooling layer
        kLRN = ESWIN_LAYER_TYPE_LRN,                          //!< LRN layer
        kSCALE = ESWIN_LAYER_TYPE_SCALE,                      //!< Scale Layer
        kBATCH_NORM = ESWIN_LAYER_TYPE_BATCH_NORM,            //!< Batch Norm Layer
        kSOFTMAX = ESWIN_LAYER_TYPE_SOFTMAX,                  //!< SoftMax layer
        kDECONVOLUTION = ESWIN_LAYER_TYPE_DECONVOLUTION,      //!< Deconvolution layer
        kCONCATENATION = ESWIN_LAYER_TYPE_CONCATENATION,      //!< Concatenation layer
        kELEMENTWISE = ESWIN_LAYER_TYPE_ELEMENTWISE,          //!< Elementwise layer
        kSLICE = ESWIN_LAYER_TYPE_SLICE,                      //!< Slice layer
        lt_kUNKNOWN = ESWIN_LAYER_TYPE_UNKNOWN,
    };
    static inline UnderlyingType max() { return 12U; }
    const char* c_str() const
    {
        const char* names[13] = {"CONVOLUTION", "FULLY_CONNECTED", "ACTIVATION", "POOLING",       "LRN",
                                 "SCALE",       "BATCHNORM",       "SOFTMAX",    "DECONVOLUTION", "CONCATENATION",
                                 "ELEMENTWISE", "SLICE",           "UNKNOWN"};
        return names[m_v];
    }
    static UnderlyingType getEnum(std::string find)
    {
        uint8_t i = 0;
        const char* names[13] = {"CONVOLUTION", "FULLY_CONNECTED", "ACTIVATION", "POOLING",       "LRN",
                                 "SCALE",       "BATCHNORM",       "SOFTMAX",    "DECONVOLUTION", "CONCATENATION",
                                 "ELEMENTWISE", "SLICE",           "UNKNOWN"};
        uint8_t match = lt_kUNKNOWN;
        for (; i <= max(); ++i) {
            if (names[i] == find) {
                match = i;
                break;
            }
        }
        return match;
    }

    ENUM_CLASS_MEMBERS(LayerType);
};
template <>
inline int EnumMax<LayerType>()
{
    return LayerType::max() + 1;
}  // used by checkers

enum ActivationType {
    kRELU = ESWIN_ACTIVATION_TYPE_RELU,
    kSIGMOID = ESWIN_ACTIVATION_TYPE_SIGMOID,
    kTANH = ESWIN_ACTIVATION_TYPE_TANH,
};
template <>
inline int EnumMax<ActivationType>()
{
    return 3;
}  // used by checkers

class PoolingType
{
 public:
    typedef uint8_t UnderlyingType;
    enum Enum {
        kMIN = ESWIN_POOLING_TYPE_MIN,         // Min over elements
        kMAX = ESWIN_POOLING_TYPE_MAX,         // Maximum over elements
        kAVERAGE = ESWIN_POOLING_TYPE_AVERAGE  // Average over elements. If the tensor is padded, the count includes the
                                               // padding
    };
    static inline UnderlyingType max() { return 2U; }
    const char* c_str() const
    {
        const char* names[3] = {"MIN", "MAX", "AVERAGE"};
        return names[m_v];
    }
    static UnderlyingType getEnum(std::string find)
    {
        uint8_t i = 0;
        const char* names[3] = {"MIN", "MAX", "AVERAGE"};
        uint8_t match = kAVERAGE;
        for (; i <= max(); ++i) {
            if (names[i] == find) {
                match = i;
                break;
            }
        }
        return match;
    }
    ENUM_CLASS_MEMBERS(PoolingType);
};
template <>
inline int EnumMax<PoolingType>()
{
    return PoolingType::max() + 1;
}  // used by checkers

enum BiasMode {
    bNONE = ESWIN_BIAS_MODE_NONE,                 //!< no bias
    bUNIFORM = ESWIN_BIAS_MODE_UNIFORM,           //!< identical coefficients across all elements of the tensor
    bCHANNEL = ESWIN_BIAS_MODE_CHANNEL,           //!< per-channel coefficients
    bm_ELEMENTWISE = ESWIN_BIAS_MODE_ELEMENTWISE  //!< elementwise coefficients
};
template <>
inline int EnumMax<BiasMode>()
{
    return 4;
}  // used by checkers

enum ScaleMode {
    sUNKNOWN = ESWIN_SCALE_MODE_UNKNOWN,           //!< unknown scale mode
    sUNIFORM = ESWIN_SCALE_MODE_UNIFORM,           //!< identical coefficients across all elements of the tensor
    sCHANNEL = ESWIN_SCALE_MODE_CHANNEL,           //!< per-channel coefficients
    sm_ELEMENTWISE = ESWIN_SCALE_MODE_ELEMENTWISE  //!< elementwise coefficients
};
template <>
inline int EnumMax<ScaleMode>()
{
    return 4;
}  // used by checkers

enum BatchNormMode {
    bnUNIFORM = ESWIN_BATCH_NORM_MODE_UNIFORM,    //!<< identical coefficients across all elements of the tensor
    bnm_CHANNEL = ESWIN_BATCH_NORM_MODE_CHANNEL,  //!< per-channel coefficients
};
template <>
inline int EnumMax<BatchNormMode>()
{
    return 2;
}  // used by checkers

enum ElementWiseOperation {
    kSUM = ESWIN_ELEMENTWISE_OPERATION_SUM,    //!< sum of the two elements
    kPROD = ESWIN_ELEMENTWISE_OPERATION_PROD,  //!< product of the two elements
    kMIN = ESWIN_ELEMENTWISE_OPERATION_MIN,    //!< minimum of the two elements
    ew_kMAX = ESWIN_ELEMENTWISE_OPERATION_MAX  //!< maximum of the two elements
};
template <>
inline int EnumMax<ElementWiseOperation>()
{
    return 3;
}  // used by checkers

class PixelMapping
{
 public:
    enum Enum {
        PITCH_LINEAR = ESWIN_PIXEL_MAPPING_PITCH_LINEAR,
    };
    typedef uint8_t UnderlyingType;

    static inline UnderlyingType max() { return 1U; }

    ENUM_CLASS_MEMBERS(PixelMapping)
};
template <>
inline int EnumMax<PixelMapping>()
{
    return PixelMapping::max() + 1;
}  // used by checkers

class PixelFormat
{
 public:
    enum Enum {
        R8 = ESWIN_PIXEL_FORMAT_R8,
        R10 = ESWIN_PIXEL_FORMAT_R10,
        R12 = ESWIN_PIXEL_FORMAT_R12,
        R16 = ESWIN_PIXEL_FORMAT_R16,
        R16_I = ESWIN_PIXEL_FORMAT_R16_I,
        R16_F = ESWIN_PIXEL_FORMAT_R16_F,
        A16B16G16R16 = ESWIN_PIXEL_FORMAT_A16B16G16R16,
        X16B16G16R16 = ESWIN_PIXEL_FORMAT_X16B16G16R16,
        A16B16G16R16_F = ESWIN_PIXEL_FORMAT_A16B16G16R16_F,
        A16Y16U16V16 = ESWIN_PIXEL_FORMAT_A16Y16U16V16,
        V16U16Y16A16 = ESWIN_PIXEL_FORMAT_V16U16Y16A16,
        A16Y16U16V16_F = ESWIN_PIXEL_FORMAT_A16Y16U16V16_F,
        A8B8G8R8 = ESWIN_PIXEL_FORMAT_A8B8G8R8,
        A8R8G8B8 = ESWIN_PIXEL_FORMAT_A8R8G8B8,
        B8G8R8A8 = ESWIN_PIXEL_FORMAT_B8G8R8A8,
        R8G8B8A8 = ESWIN_PIXEL_FORMAT_R8G8B8A8,
        X8B8G8R8 = ESWIN_PIXEL_FORMAT_X8B8G8R8,
        X8R8G8B8 = ESWIN_PIXEL_FORMAT_X8R8G8B8,
        B8G8R8X8 = ESWIN_PIXEL_FORMAT_B8G8R8X8,
        R8G8B8X8 = ESWIN_PIXEL_FORMAT_R8G8B8X8,
        A2B10G10R10 = ESWIN_PIXEL_FORMAT_A2B10G10R10,
        A2R10G10B10 = ESWIN_PIXEL_FORMAT_A2R10G10B10,
        B10G10R10A2 = ESWIN_PIXEL_FORMAT_B10G10R10A2,
        R10G10B10A2 = ESWIN_PIXEL_FORMAT_R10G10B10A2,
        A2Y10U10V10 = ESWIN_PIXEL_FORMAT_A2Y10U10V10,
        V10U10Y10A2 = ESWIN_PIXEL_FORMAT_V10U10Y10A2,
        A8Y8U8V8 = ESWIN_PIXEL_FORMAT_A8Y8U8V8,
        V8U8Y8A8 = ESWIN_PIXEL_FORMAT_V8U8Y8A8,
        Y8___U8V8_N444 = ESWIN_PIXEL_FORMAT_Y8___U8V8_N444,
        Y8___V8U8_N444 = ESWIN_PIXEL_FORMAT_Y8___V8U8_N444,
        Y10___U10V10_N444 = ESWIN_PIXEL_FORMAT_Y10___U10V10_N444,
        Y10___V10U10_N444 = ESWIN_PIXEL_FORMAT_Y10___V10U10_N444,
        Y12___U12V12_N444 = ESWIN_PIXEL_FORMAT_Y12___U12V12_N444,
        Y12___V12U12_N444 = ESWIN_PIXEL_FORMAT_Y12___V12U12_N444,
        Y16___U16V16_N444 = ESWIN_PIXEL_FORMAT_Y16___U16V16_N444,
        Y16___V16U16_N444 = ESWIN_PIXEL_FORMAT_Y16___V16U16_N444,
        FEATURE = ESWIN_PIXEL_FORMAT_FEATURE,
        FEATURE_X8 = ESWIN_PIXEL_FORMAT_FEATURE_X8,
        UNKNOWN = ESWIN_PIXEL_FORMAT_UNKNOWN,
    };
    typedef uint8_t UnderlyingType;
    static inline UnderlyingType max() { return 38U; }
    const char* c_str() const
    {
        const char* names[39] = {"R8",
                                 "R10",
                                 "R12",
                                 "R16",
                                 "R16_I",
                                 "R16_F",
                                 "A16B16G16R16",
                                 "X16B16G16R16",
                                 "A16B16G16R16_F",
                                 "A16Y16U16V16",
                                 "V16U16Y16A16",
                                 "A16Y16U16V16_F",
                                 "A8B8G8R8",
                                 "A8R8G8B8",
                                 "B8G8R8A8",
                                 "R8G8B8A8",
                                 "X8B8G8R8",
                                 "X8R8G8B8",
                                 "B8G8R8X8",
                                 "R8G8B8X8",
                                 "A2B10G10R10",
                                 "A2R10G10B10",
                                 "B10G10R10A2",
                                 "R10G10B10A2",
                                 "A2Y10U10V10",
                                 "V10U10Y10A2",
                                 "A8Y8U8V8",
                                 "V8U8Y8A8",
                                 "Y8___U8V8_N444",
                                 "Y8___V8U8_N444",
                                 "Y10___U10V10_N444",
                                 "Y10___V10U10_N444",
                                 "Y12___U12V12_N444",
                                 "Y12___V12U12_N444",
                                 "Y16___U16V16_N444",
                                 "Y16___V16U16_N444",
                                 "FEATURE",
                                 "FEATURE_X8",
                                 "UNKNOWN"};
        return names[m_v];
    }
    static UnderlyingType getEnum(std::string find)
    {
        uint8_t i = 0;
        const char* names[39] = {"R8",
                                 "R10",
                                 "R12",
                                 "R16",
                                 "R16_I",
                                 "R16_F",
                                 "A16B16G16R16",
                                 "X16B16G16R16",
                                 "A16B16G16R16_F",
                                 "A16Y16U16V16",
                                 "V16U16Y16A16",
                                 "A16Y16U16V16_F",
                                 "A8B8G8R8",
                                 "A8R8G8B8",
                                 "B8G8R8A8",
                                 "R8G8B8A8",
                                 "X8B8G8R8",
                                 "X8R8G8B8",
                                 "B8G8R8X8",
                                 "R8G8B8X8",
                                 "A2B10G10R10",
                                 "A2R10G10B10",
                                 "B10G10R10A2",
                                 "R10G10B10A2",
                                 "A2Y10U10V10",
                                 "V10U10Y10A2",
                                 "A8Y8U8V8",
                                 "V8U8Y8A8",
                                 "Y8___U8V8_N444",
                                 "Y8___V8U8_N444",
                                 "Y10___U10V10_N444",
                                 "Y10___V10U10_N444",
                                 "Y12___U12V12_N444",
                                 "Y12___V12U12_N444",
                                 "Y16___U16V16_N444",
                                 "Y16___V16U16_N444",
                                 "FEATURE",
                                 "FEATURE_X8",
                                 "UNKNOWN"};
        uint8_t match = UNKNOWN;
        for (; i <= max(); ++i) {
            if (names[i] == find) {
                match = i;
                break;
            }
        }
        return match;
    }

    ENUM_CLASS_MEMBERS(PixelFormat);
};
template <>
inline int EnumMax<PixelFormat::Enum>()
{
    return PixelFormat::max() + 1;
}  // used by checkers

}  // namespace eswin

#undef ENUM_CLASS_MEMBERS

#endif  // ESWIN_HOST_RUNTIME_INCLUDE_ITYPE_H_
