// Copyright (c) 2023 ESWIN Authors. All Rights Reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#ifndef ESWIN_HOST_RUNTIME_INCLUDE_LOADABLE_H_
#define ESWIN_HOST_RUNTIME_INCLUDE_LOADABLE_H_

#include <map>
#include <vector>
#include <string>
#include <iLoadable.h>
#include "privType.h"
#include "loadable_generated.h"

namespace eswin
{

namespace priv
{

class Loadable;

class LoadableFactory
{
 public:
    typedef PrivPair<ILoadable *, Loadable *> LoadablePrivPair;

    static LoadablePrivPair newLoadable();
    static void deleteLoadable(ILoadable *loadable);

    static Loadable *priv(ILoadable *);
    static ILoadable *i(Loadable *);
    static ILoadable *self(void *s);
    static ILoadable *deserializeLoadable(uint8_t *);

 protected:
    static BiMap<ILoadable *, Loadable *> s_priv;
    static BiMap<void *, ILoadable *> s_self;
};

class Loadable : public ILoadable
{
 public:  // externally facing
    virtual std::string getName() const;

    virtual int getNumMemoryListEntries() const;
    virtual MemoryListEntry getMemoryListEntry(uint16_t pool_id) const;

    virtual int getNumEventListEntries() const;
    virtual EventListEntry getEventListEntry(uint16_t event_id) const;

    virtual int getNumSubModelListEntries() const;
    virtual SubModelListEntry getSubModelListEntry(uint16_t task_id) const;

    virtual int getNumSubmitListEntries() const;
    virtual SubmitListEntry getSubmitListEntry(uint16_t submit_id) const;

    virtual int getNumAddressListEntries() const;
    virtual AddressListEntry getAddressListEntry(uint16_t address_list_index) const;

    virtual int getNumTensorDescListEntries() const;
    virtual TensorDescListEntry getTensorDescListEntry(uint16_t tensor_desc_list_index) const;

    virtual int32_t getNetworkDataType(DataType::UnderlyingType *) const;

    virtual int32_t getNumInputTensors(int *) const;
    virtual int32_t getInputTensorDesc(uint16_t id, TensorDescListEntry *) const;

    virtual int32_t getNumOutputTensors(int *) const;
    virtual int32_t getOutputTensorDesc(uint16_t id, TensorDescListEntry *) const;

    virtual int getNumRelocEntries() const;
    virtual RelocEntry getRelocEntry(uint16_t i) const;

 public:  // internally facing
    virtual int setSymbolContent(std::string name, const ILoadable::Blob &, uint8_t *data);
    virtual bool getSymbolContent(std::string name, ILoadable::Blob &, uint8_t *&);

    virtual uint16_t getFactoryType() const;

    void setMemoryListEntries(const std::vector<MemoryListEntry> &);
    void setEventListEntries(const std::vector<EventListEntry> &);
    void setSubModelListEntries(const std::vector<SubModelListEntry> &);
    void setSubmitListEntries(const std::vector<SubmitListEntry> &);
    void setAddressListEntries(const std::vector<AddressListEntry> &);
    void setTensorDescListEntries(const std::vector<TensorDescListEntry> &);
    void setRelocEntries(const std::vector<RelocEntry> &);

    std::vector<SubModelListEntry> getSubModelListEntries() const;
    const std::vector<SubmitListEntry> &getSubmitListEntries() const;
    const std::vector<MemoryListEntry> &getMemoryListEntries() const;
    const std::vector<AddressListEntry> &getAddressListEntries() const;
    const std::vector<EventListEntry> &getEventListEntries() const;
    const std::vector<TensorDescListEntry> &getTensorDescListEntries() const;
    const std::vector<RelocEntry> &getRelocEntries() const;

    Loadable();
    virtual ~Loadable();

    virtual bool serialize();
    virtual int32_t getSerializedData(uint8_t *buffer);
    virtual int32_t getSerializedDataSize(uint64_t *size);
    virtual bool deserializeFrom(uint8_t *);

    struct Symbol {
        std::string name;
        ILoadable::Interface interface;
        uint32_t subInterface;
        ILoadable::Version version;
        uint64_t size;
        uint8_t *data;
        Symbol() {}
    };

    inline bool debugSymbolContent() { return false; }

 protected:
    friend class Runtime;
    std::map<std::string, Symbol> mSymbols;
    std::vector<MemoryListEntry> mMemoryListEntries;
    std::vector<SubModelListEntry> mSubModelListEntries;
    std::vector<SubmitListEntry> mSubmitListEntries;
    std::vector<EventListEntry> mEventListEntries;
    std::vector<AddressListEntry> mAddressListEntries;
    std::vector<TensorDescListEntry> mTensorDescListEntries;
    std::vector<RelocEntry> mRelocEntries;

    std::string mName;

 private:
    flatbuffers::FlatBufferBuilder mFbb;
};

}  // namespace priv

}  // namespace eswin

#endif  // ESWIN_HOST_RUNTIME_INCLUDE_LOADABLE_H_
