// Copyright (c) 2023 ESWIN Authors. All Rights Reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#ifndef ESWIN_HOST_RUNTIME_INCLUDE_MODEL_H_
#define ESWIN_HOST_RUNTIME_INCLUDE_MODEL_H_

#include "hetero_env.h"
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <sys/mman.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <unistd.h>
#include <cstdio>
#include <cstring>
#include <sstream>
#include <map>
#include <list>
#include <vector>
#include "es_npu_error.h"
#include "loadable.h"
#include "modelSub.h"
#include "core.h"
#include "task.h"
#include "hetero_common.h"
#include "hetero_processor.h"
#include "hetero_ioctl.h"
#include "npuMemPool.h"
#include "es_dsp_internal.h"
#include "hwProcessorMgr.h"
#include <string>
#include <utility>
#include <memory>
#include <unordered_map>
#if NPU_DEV_SIM != NPU_ESIM_TOOL
#include "performance.h"
#endif

namespace eswin
{
// Invalid number of operator index
#define INVALID_OP_INDEX (-1)

#define NPU_DMABUF_ALIGN_SIZE 0x200000UL  // 2M
#define ALIGNMENT_UP(size, align) (((size) + ((align)-1)) & (~((align)-1)))
#define NPU_MODEL_ALLOC_ZONE_NID0 "mmz_nid_0_part_0"
#define NPU_MODEL_ALLOC_ZONE_NID1 "mmz_nid_1_part_0"

class DspProcessor;
class NpuProcessor;

namespace priv
{

enum {
    SUBMODEL_KMD = 0,
    SUBMODEL_UMD,
};

class TensorDesc : public ILoadable::TensorDescListEntry
{
 public:
    TensorDesc();
    explicit TensorDesc(const eswin::ILoadable::TensorDescListEntry &e);
    NPU_TENSOR_S bindTensorDesc() const;
};

class SubModel;
class UmdSubModel;
class KmdSubModel;

class esModel
{
 public:
    /**
     * @brief            read mode buffer
     *
     * read mode buffer from file.
     * @param modelPath  the model file path
     * @param modelBuf   binary cache of the model
     * @return           0 on success and negative on error
     */
    ES_S32 readModelToMem(const char *modelPath, char **modelBuf);

    /**
     * @brief            instantiate a model
     *
     * deserialize the loadabel struct from modelBuf,then parse out the mMemoryListEntry/
     * mAddressListEntry/mSubModelListEntry.
     * then can obtain the corresponding types from the LisEntry.
     * @param modelBuf   binary cache of the model
     * @return           0 on success and negative on error
     */
    ES_S32 instantiateModelFromMem(char *modelBuf, NPU_MODEL_TYPE_E modelType);

    /**
     * @brief            set model id
     *
     * set model id.
     * @param modelId  model id
     */
    void setModelId(uint32_t modelId);

    /**
     * @brief            get model id
     *
     * get model id.
     * @return  model id
     */
    uint32_t getModelId(void);

    /**
     * @brief            get pointer of model recorder struct
     *
     * get pointer of model recorder struct.
     * @return  pointer of ModelRec
     */
    modelRec_t *getModelRec();
#if NPU_DEV_SIM == NPU_ESIM_TOOL
    modelRec_t *getConcatModelRec();
    int32_t concatSubModels();
#endif

    /**
     * @brief            submit model
     *
     * call the submitSubModel function of each sub model
     * @return           0 on success and negative on error
     */
    ES_S32 submitModel();

    /**
     * @brief            release model
     *
     * call the releaseSubModel function of each sub model
     * @return           0 on success and negative on error
     */
    ES_S32 releaseModel(NPU_MODEL_TYPE_E modelType);

    /**
     * @brief            get umd submodel by index.
     *
     * get umd submodel.
     * @param index      the index of umd submodel
     * @return           the pointer of UmdSubModel, nullptr if failed
     */
    UmdSubModel *getUmdSubModel(int16_t index);

    /**
     * @brief            get kmd submodel by umd submodel
     *
     * get kmd submodel corresponding to the umd submodel.
     * @param umd        the pointer of umd submodel
     * @return           the pointer of kmd submodel, nullptr if failed
     */
    const KmdSubModel *getKmdSubModel(const UmdSubModel *umd);

    /**
     * @brief            get total number of input tesnsor
     *
     * @param inputs     total number of input tesnsor
     * @return           0 on success and negative on error
     */
    ES_S32 getInputTensors(int *inputs);
    ES_S32 getOutputTensors(int *outputs);

    /**
     * @brief            get input tensor describe
     *
     * get input tensor describe
     * @param id   index of tensor, the value  starts from 0
     * @param td   pointer of tensor describe
     * @return           0 on success and negative on error
     */
    ES_S32 getInputTensorDesc(int id, NPU_TENSOR_S *td);
    ES_S32 getOutputTensorDesc(int id, NPU_TENSOR_S *td);
    int32_t getNDims();
    std::string &getDumpPath();
    ES_S32 enablePerf();
    int32_t disablePerf();
    int32_t dumpPerformanceData(const char *dir);
    void enableDump();
    void disableDump();
    bool isDumpEnable();
    int32_t configDumpInfo(const char *dumpPath, NPU_DUMP_CONF_S *dumpConf, uint32_t dumpConfNum);
    void incrementRunTimes();
    int32_t ddrBandwidth(uint64_t *readSize, uint64_t *writeSize);
    static uint32_t generateModelId();
    uint64_t getMemListAlignedSize() { return mMemListAlignedSize; }

    class MemoryBindIdLTCompare  // helper comparator
    {
     public:
        bool operator()(Memory *const &i, Memory *const &j)
        {
            IOD na;
            return i->bindId(na) < j->bindId(na);
        }
    };

    class MemoryIdBindIdIs  // helper predicate
    {
     public:
        explicit MemoryIdBindIdIs(int findId) : mfindId(findId) {}
        bool operator()(Memory *&mem) const
        {
            IOD na;
            return mfindId == mem->bindId(na);
        }

     protected:
        int mfindId;
    };

 private:
    int initBindableMemory();
    int loadMemory(Loadable *l, Memory *memory, NPU_MODEL_TYPE_E modelType);
    void unloadMemory(Memory *memory, NPU_MODEL_TYPE_E modelType);
    int getMemoryFromBindId(IOD w, int id, Memory *&boundMem);
    int32_t loadSramMemory(int32_t *sramFd, void **virtAddr, uint64_t size);
    int32_t unLoadSramMemory(int32_t sramFd, void *virtAddr, uint64_t size);
    std::string getModelName(const char *path);

    std::vector<eswin::priv::Memory> mMemorys;
    std::vector<eswin::priv::Address> mAddresses;
    std::vector<eswin::priv::TensorDesc> mTensorDesc;
    std::vector<eswin::priv::SubModel *> mSubModels;

    std::vector<eswin::ILoadable::MemoryListEntry> mMemoryListEntry;
    std::vector<eswin::ILoadable::AddressListEntry> mAddressListEntry;
    std::vector<eswin::ILoadable::TensorDescListEntry> mTensorDescListEntry;
    std::vector<eswin::ILoadable::SubModelListEntry> mSubModelListEntry;

    std::vector<std::vector<Memory *>> mBindableMemory;  // indexed on [iod][bind_id]

    uint32_t mModelId;

    uint64_t mMemListAlignedSize;

    /* the description of shared memory between user space and kernel space */
    modelRec_t mModelRec;
    static uint32_t gModelId;
    bool mDumpEnable;
    std::string mDumpPath;
    std::string mModelName;
    std::atomic<uint32_t> mRuntimes;
#if NPU_DEV_SIM == NPU_ESIM_TOOL
    char *mConcatBuf;
    struct dla_network_desc *mConcatNetwork;
    modelRec_t mConcatModelRec;
    struct dla_network_desc concatNetworkDesc;
#endif
};

class SubModel
{
 public:
    virtual ~SubModel() {}
    void dumpModel() const;
    void loadModelDesc();
    const dataDesc_t *getDataDesc() const { return &mModelDesc; }
    int getSubModelType() const { return mEntry->interface; }

    /**
     * @brief            submit submodel
     *
     * parse the structure of network/common/surface/operation from addressList.
     * if subModel is npu, submit to kmd.
     * if subModel if umd, load dsp operators.
     * @return           positive on success and negative on error
     */
    virtual int32_t submitSubModel(esModel *model, ILoadable::SubModelListEntry *entry) = 0;
#if NPU_DEV_SIM == NPU_ESIM_TOOL
    virtual int32_t getSubModelTask(esModel *model, ILoadable::SubModelListEntry *entry) = 0;
#endif

    /**
     * @brief            unload submodel
     *
     * release the resources requested during submitSubModel.
     * @return           0 on success and negative on error
     */
    virtual int32_t releaseSubModel() = 0;

 protected:
    ILoadable::SubModelListEntry *mEntry;
    eswin::priv::esModel *mModel;
    std::vector<int16_t> mOpDumpList;
    dataDesc_t mModelDesc;
};

class UmdSubModel : public SubModel
{
 public:
    UmdSubModel();

    /**
     * @brief           get corresponding kmd submodel id.
     *
     * Umd always has a corresponding kmd, get the id of kmd corresponding to itself.
     * @return the id of kmd
     */
    int16_t getKmdId() const;

    /**
     * @brief           fill dependency array of umd submodel
     *
     * fill dependency array of umd submodel with mDependency member in class.
     * @param dependency pointer of dependency array
     */
    void fillDependency(uint8_t *dependency) const;

    /**
     * @brief           get op index of event op
     *
     * get op index of event sink op by event id.
     * @param eventId the event id
     * @return op index
     */
    int16_t getEventSinkOpIndex(int16_t eventId) const;

    /**
     * @brief           get op index of event op
     *
     * get op index of event source op by event id.
     * @param eventId the event id
     * @return op index
     */
    int16_t getEventSourceOpIndex(int16_t eventId) const;

    /**
     * @brief            get struct of OpTaskDesc
     *
     * Get pointer of OpTaskDesc which is used for task scheduler.
     * @return the pointer of OpTaskDesc if succeed, null if failed
     */
    opTaskDesc_t *getOpTaskDesc(void) const;

    /**
     * @brief            get dsp op list
     *
     * Get reference of the dsp op list.
     * @param devType the device type of dsp.
     * @return the reference of the dsp op list.
     */
    std::vector<DSP_TASK_DESC_S> &getDspOpList(DEVICE_TYPE_E devType);

    /**
     * @brief            update dsp op io tensor
     *
     * Update io tensor for dsp op when submit task.
     * @param dspTask the dsp task will be update.
     * @param npuTask the io tensor was saved in it.
     * @param opTask specify the op will be update.
     * @return the reference of the dsp op list.
     */
    void updateDspOpIoTensor(ES_DSP_TASK_S *dspTask, npuTask_t *npuTask, opTaskDesc_t *opTask);

    /**
     * @brief            update dsp op list io tensor
     *
     * Update io tensor for all dsp op when submit dsp model task.
     * @param devType the device type of dsp.
     * @return the io tensor was saved in it.
     */
    void updateDspOpListIoTensor(DEVICE_TYPE_E devType, npuTask_t *npuTask);

    /**
     * Load mNetwork and mTaskDesc struct. Load mDependencySave. Load dsp operator.
     * @param model the pointer of esModel
     * @param entry the pointer of SubModelListEntry
     * @return 0 if succeed, else failed
     */
    int32_t submitSubModel(esModel *model, ILoadable::SubModelListEntry *entry);
    int32_t setDumpOp(const char *dumpList);
#if NPU_DEV_SIM == NPU_ESIM_TOOL
    int32_t getSubModelTask(esModel *model, ILoadable::SubModelListEntry *entry);
#endif

    /**
     * Release resource.
     */
    int32_t releaseSubModel();

 private:
    /**
     * analyzed from the umdSubModel's common_desc.
     */
    uint8_t *mDependecy;
    uint32_t mNumEventSinkOp;
    uint32_t mNumEventSourceOp;
    std::vector<int16_t> mEventSinkOpIndex;
    std::vector<int16_t> mEventSourceOpIndex;

    opTaskDesc_t *mOpTaskDesc;
    std::unordered_map<int8_t, std::shared_ptr<DspProcessor>> mDspProcessors;
    std::unordered_map<std::string, std::pair<int8_t, int32_t>> mDspOpHandleMap;
    std::unordered_map<DEVICE_TYPE_E, std::vector<DSP_TASK_DESC_S>> mDspOplistMap;
    std::unordered_map<DEVICE_TYPE_E, std::vector<int32_t>> mDspOpIndexMap;

    int32_t loadDependency();
    void loadEventOpIndex();
    int32_t loadOpTaskDesc();
    int32_t loadDspOperator();
    void loadEsDspTask(uint16_t index);
    void loadEsHaeTask(uint16_t index);
    void loadDspOpList();
};

class KmdSubModel : public SubModel
{
 public:
    int32_t submitSubModel(esModel *model, ILoadable::SubModelListEntry *entry);
    int32_t releaseSubModel();
    uint16_t getModelId() const;
    int32_t setDumpInfo(const char *dumpPath, const char *dumpList);
    const kmd_dump_info_t *getDumpInfo() const;
#if NPU_DEV_SIM == NPU_ESIM_TOOL
    int32_t getSubModelTask(esModel *model, ILoadable::SubModelListEntry *entry);
#endif

 private:
    /**
     *  mSubModelId obtained from kmd
     */
    uint16_t mSubModelId;
    std::shared_ptr<NpuProcessor> mNpuProcessor;
    kmd_dump_info_t mDumpInfo;
};

}  // namespace priv
}  // namespace eswin

#endif  // ESWIN_HOST_RUNTIME_INCLUDE_MODEL_H_
