// Copyright (c) 2023 ESWIN Authors. All Rights Reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#ifndef ESWIN_HOST_RUNTIME_INCLUDE_MODELMANAGE_H_
#define ESWIN_HOST_RUNTIME_INCLUDE_MODELMANAGE_H_
#include "hetero_env.h"
#include <vector>
#include <unordered_map>
#include <memory>
#include <string>
#include "model.h"
#include "es_npu_types.h"

namespace eswin
{

typedef struct _modelInfo {
    NPU_MODEL_INFO_S info;
    uint32_t batchNum;
} modelInfo_s;

typedef struct _compositeModelInfo {
    uint32_t modelNums;
    modelInfo_s modelsInfo[MAX_MODELS_OF_SET];
} compositeModelInfo_t;

class esModelManager
{
 public:
    esModelManager() { pthread_mutex_init(&mMutex, NULL); }
    ~esModelManager() { pthread_mutex_destroy(&mMutex); }
    int32_t loadModel(uint32_t *modelId, const char *modelPath, NPU_MODEL_TYPE_E modelType);
    int32_t loadCompositeModel(uint32_t *modelId, const char *modelPath);
    int32_t unloadModel(uint32_t modelId, NPU_MODEL_TYPE_E modelType);
    int32_t unloadCompositeModel(uint32_t modelId);
    int32_t getModel(uint32_t modelId, eswin::priv::esModel **model);
    int32_t getCompositeModelInfo(uint32_t modelId, compositeModelInfo_t *compositeModelInfo);
    int32_t setCompositeModelAttr(uint32_t modelId, NPU_FLEXIBLE_TASK_ATTR_S *taskAttr);
    int32_t getCompositeModelAttr(uint32_t modelId, NPU_FLEXIBLE_TASK_ATTR_S *taskAttr);
    modelRec_t *getModelRec(uint32_t modelId);
    int32_t getMemBlock(uint32_t modelId, NPU_TASK_MEM_S *taskMem);
    int32_t releaseMemBlock(uint32_t modelId, NPU_TASK_MEM_S *taskMem);
    uint32_t queryModelHwMask(uint32_t modelId);
#if NPU_DEV_SIM == NPU_ESIM_TOOL
    modelRec_t *getConcatModelRec(uint32_t modelId);
#endif

 private:
    std::unordered_map<uint32_t, priv::esModel *> mModelMap;
    std::unordered_map<uint32_t, compositeModelInfo_t> mCompositeModelMap;
    std::unordered_map<uint32_t, std::shared_ptr<NpuMemPool>> mMemPoolMap;
    std::unordered_map<uint32_t, NPU_FLEXIBLE_TASK_ATTR_S> mModelTaskAttr;
    std::unordered_map<uint32_t, uint32_t> mModelHwMaskMap;
    pthread_mutex_t mMutex;
};

modelRec_t *getModelRec(uint32_t modelId);
#if NPU_DEV_SIM == NPU_ESIM_TOOL
modelRec_t *getConcatModelRec(uint32_t modelId);
#endif
int32_t loadModelFromFile(uint32_t *modelId, const char *modelPath);
int32_t loadCompositeModel(uint32_t *modelId, const char *modelPath);
int32_t unloadModel(uint32_t modelId);
int32_t unloadCompositeModel(uint32_t modelId);
int32_t getMemFromModelManager(uint32_t modelId, NPU_TASK_MEM_S *taskMem);
int32_t releaseMemToModelManager(uint32_t modelId, NPU_TASK_MEM_S *taskMem);
int32_t getNumInputTensors(uint32_t modelId, int32_t *inputTensors);
int32_t getNumOutputTensors(uint32_t modelId, int32_t *outputTensors);
int32_t getInputTensorDesc(uint32_t modelId, int32_t tensorId, NPU_TENSOR_S *tensor);
int32_t getOutputTensorDesc(uint32_t modelId, int32_t tensorId, NPU_TENSOR_S *tensor);
int32_t getModel(uint32_t modelId, eswin::priv::esModel **model);
int32_t getCompositeModelInfo(uint32_t modelId, NPU_COMPOSITE_MODEL_INFO_S *compositeModelInfo);
int32_t innerGetCompositeModelInfo(uint32_t modelId, compositeModelInfo_t *compositeModelInfo);
int32_t getCompositeModelAttr(uint32_t modelId, NPU_FLEXIBLE_TASK_ATTR_S *attr);
int32_t setCompositeModelAttr(uint32_t modelId, NPU_FLEXIBLE_TASK_ATTR_S *attr);
uint32_t queryModelHwMask(uint32_t modelId);
int32_t configDumpInfo(uint32_t modelId, const char *dumpPath, NPU_DUMP_CONF_S *dumpConf, uint32_t dumpConfNum);
int32_t ddrBandwidth(uint32_t modelId, uint64_t *readSize, uint64_t *writeSize);
int32_t enableDump(uint32_t modelId);
int32_t disableDump(uint32_t modelId);
int32_t dumpOpStatus(uint32_t modelId);
int32_t enablePerformance(ES_U32 modelId);
int32_t disablePerformance(ES_U32 modelId);
int32_t dumpPerformanceData(ES_U32 modelId, const ES_CHAR *dir);
NPU_MODEL_TYPE_E convertToModelType(const std::string &type);

}  // namespace eswin

#endif  // ESWIN_HOST_RUNTIME_INCLUDE_MODELMANAGE_H_
