// Copyright (c) 2023 ESWIN Authors. All Rights Reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#ifndef ESWIN_HOST_RUNTIME_INCLUDE_MODELSUB_H_
#define ESWIN_HOST_RUNTIME_INCLUDE_MODELSUB_H_

#include <string>
#include <map>
#include <vector>
#include <memory>
#include <algorithm>
#include <functional>
#include <semaphore.h>
#include "privType.h"
#include "iLoadable.h"

namespace eswin
{
namespace priv
{
class Memory
{
 public:
    Memory() : Fd(-1), pVirtAddr(0) {}
    Memory(const ILoadable::MemoryListEntry &e) : Fd(-1), pVirtAddr(0), mEntry(e) {}
    Memory(const Memory &o) : Fd(o.Fd), pVirtAddr(0), mEntry(o.mEntry) {}
    inline uint16_t id() { return mEntry.id; }
    inline uint64_t size() { return mEntry.size; }
    inline uint32_t alignment() { return mEntry.alignment; }
    inline uint8_t domain() { return mEntry.domain; }
    inline bool bound() { return Fd != -1; }
    inline uint8_t flags() { return mEntry.flags; }
    inline void setFd(int fd) { Fd = fd; }
    inline int getFd() const { return Fd; }
    inline void setVirtAddr(void *addr) { pVirtAddr = addr; }
    inline void *getVirtAddr() const { return pVirtAddr; }
    inline std::vector<std::string> &contents() { return mEntry.contents; }
    inline std::vector<uint64_t> &offsets() { return mEntry.offsets; }
    inline int inputBindId() const
    {
        if (mEntry.flags & mEntry.flags_input()) {
            return (int)mEntry.bind_id;
        }
        return -1;
    }
    inline int outputBindId() const
    {
        if (mEntry.flags & mEntry.flags_output()) {
            return (int)mEntry.bind_id;
        }
        return -1;
    }
    inline bool bindable() const { return !!(mEntry.flags & (mEntry.flags_input() | mEntry.flags_output())); }
    inline int bindId(IOD &which) const
    {
        // there should be only one valid.  but if not take in order of
        // input, output
        if (mEntry.flags & mEntry.flags_input()) {
            which = IOD_Input;
            return (int)mEntry.bind_id;
        } else if (mEntry.flags & mEntry.flags_output()) {
            which = IOD_Output;
            return (int)mEntry.bind_id;
        }
        which = IOD_Max;
        return -1;
    };
    inline int tensorDescId() const
    {
        if (mEntry.flags & (mEntry.flags_input() | mEntry.flags_output())) {
            return (int)mEntry.tensor_desc_id;
        }
        return -1;
    }

 protected:
    int Fd;
    void *pVirtAddr;
    ILoadable::MemoryListEntry mEntry;
};

class Event
{
 public:
    uint16_t id() const { return mEntry.id; }
    uint8_t op() const { return mEntry.op; }
    uint16_t target() const { return mEntry.target; }
    uint32_t val() const { return mEntry.val; }

 protected:
    ILoadable::EventListEntry mEntry;
};

class Address
{
 public:
    Address() {}
    Address(const ILoadable::AddressListEntry &e) : mEntry(e) {}
    Address(const Address &o) : mEntry(o.mEntry) {}
    uint16_t id() const { return mEntry.id; }
    uint16_t mem_id() const { return mEntry.mem_id; }
    uint64_t offset() const { return mEntry.offset; }

 public:
    ILoadable::AddressListEntry mEntry;
};

}  // namespace priv
}  // namespace eswin

#endif  // ESWIN_HOST_RUNTIME_INCLUDE_MODELSUB_H_
