// Copyright (c) 2023 ESWIN Authors. All Rights Reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#ifndef ESWIN_HOST_RUNTIME_INCLUDE_TASKSCHEDULER_H_
#define ESWIN_HOST_RUNTIME_INCLUDE_TASKSCHEDULER_H_

#include <map>
#include <atomic>
#include <memory>
#include <vector>
#include <thread>
#include <string>
#include <cstdint>
#include <unordered_map>
#include "esMutex.h"
#include "core.h"
#include "model.h"

#define MAX_NUM_UMD_OP 1024
namespace eswin
{
class TaskScheduler
{
 public:
    explicit TaskScheduler(npuContext_t *ctx);
    ~TaskScheduler();
    void startSchedule();
    void stopSchedule();
    priv::esModel *getModel();
    npuTask_t *getNpuTask();
    static int32_t complete(void *arg);
    static void dspComplete(void *arg, int32_t state);

 private:
    void schedulerLoop();
    DEVICE_TYPE_E prepareTask();
    int32_t submitTask(DEVICE_TYPE_E device);
    int32_t umdSubmitTask(uint32_t umdSubmodelId);
    int32_t kmdSubmitTask(uint16_t kmdSubmodelId, const kmd_dump_info_t *dumpInfo, bool isDump);
    void waitTaskComplete();
    int32_t handleCompletedTask(DEVICE_TYPE_E device);
    void waitTaskCompletedThread();
    void submitTaskThread();

    // helper function
    bool schedulerExecute();
    void handleFailure(const std::string &message);
    static int32_t writeBufToFile(const char *path, void *buf, uint32_t size);
    static int32_t dumpUmdOp(opTaskDesc_t *opTask, bool isOutput);

    // Attributes
 private:
    std::atomic<bool> mRunning;
    std::thread mScheThread;
    std::thread mTaskCompletedThread;
    std::thread mSubmitTaskThread;

    /**
     * used by the task, restored to mDependecySave before new task is executed.
     */
    uint8_t mUmdDependecy[MAX_NUM_UMD_OP];  // 1024
    npuContext_t *mCtx;
    EsMutex esMutex;
    std::mutex mWaitCompletedMutex;
    std::condition_variable mWaitCompletedCond;

    std::atomic<bool> mIsAsyncSubmitTask;
    std::mutex mSubmitTaskMutex;
    std::condition_variable mAsyncTaskCond;

    std::unordered_map<DEVICE_TYPE_E, npuTask_t *> mTaskMap;
    std::unordered_map<DEVICE_TYPE_E, priv::esModel *> mModelMap;

    union {
        struct {
            int32_t dsp0 : 1;  // 1: hw is idle or task completed or hw is accessed.
            int32_t dsp1 : 1;
            int32_t dsp2 : 1;
            int32_t dsp3 : 1;
            int32_t hae : 1;
            int32_t gpu : 1;
            int32_t npu : 1;
            int32_t reserve : 25;
        };
        std::atomic<int32_t> value;
#define HW_MASK 0x4f
#define DSP_MASK 0xf
#define NPU_MASK 0x40
    } mHwStatus, mHwCompleteStatus, mHwAccessed, mHwCustom;
};
}  // namespace eswin

#endif  // ESWIN_HOST_RUNTIME_INCLUDE_TASKSCHEDULER_H_
