// Copyright (c) 2023 ESWIN Authors. All Rights Reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#ifndef ESWIN_HOST_RUNTIME_INCLUDE_UTILS_H_
#define ESWIN_HOST_RUNTIME_INCLUDE_UTILS_H_

#include <stddef.h>
#include <dirent.h>
#include "es_npu_error.h"
#include "es_npu_types.h"
#include "stdint.h"

#define ESWIN_OPEN_WRITE (0x1)
#define ESWIN_OPEN_READ (0x2)
#define ESWIN_OPEN_CREATE (0x4)
#define ESWIN_OPEN_APPEND (0x8)

/*
 * Files and directory structures.
 */
typedef enum {
    esFileType_Unknown = 0,
    esFileType_File,
    esFileType_Directory,
    esFileType_Fifo,
    esFileType_CharacterDevice,
    esFileType_BlockDevice,

    esFileType_Force32 = 0x7FFFFFFF
} esFileType;

struct esStatTypeRec {
    uint64_t size;
    esFileType type;
    uint64_t mtime;
};
typedef struct esStatTypeRec esStatType;

enum esSeek {
    esSeek_Set,
    esSeek_Cur,
    esSeek_End
};
typedef enum esSeek esSeekEnum;

struct esFileRec {
    int fd;
};
typedef struct esFileRec esFile;
typedef struct esFileRec *esFileHandle;

struct esDirRec {
    DIR *dir;
};
typedef struct esDirRec esDir;
typedef struct esDirRec *esDirHandle;

#ifdef __cplusplus
extern "C" {
#endif

/*
 * General OS related calls.
 */
void *esAlloc(size_t size);
void esFree(void *ptr);

void esDebugPrintf(const char *format, ...);

uint32_t esGetTimeMS(void);
void esSleepMS(uint32_t msec);

/*
 * File and directory operations
 */
int32_t esStat(const char *filename, esStatType *stat);
int32_t esMkdir(char *dirname);
int32_t esFremove(const char *filename);

int32_t esFopen(const char *path, uint32_t flags, esFileHandle *file);
void esFclose(esFileHandle stream);
int32_t esFwrite(esFileHandle stream, const void *ptr, size_t size);
int32_t esFread(esFileHandle stream, void *ptr, size_t size, size_t *bytes);
int32_t esFseek(esFileHandle file, int64_t offset, esSeekEnum whence);
int32_t esFstat(esFileHandle file, esStatType *stat);
uint64_t esStatGetSize(esStatType *stat);
int32_t esFgetc(esFileHandle stream, uint8_t *c);
void esMemset(void *s, uint8_t c, size_t size);

int32_t esOpendir(const char *path, esDirHandle *dir);
int32_t esReaddir(esDirHandle dir, char *name, size_t size);
void esClosedir(esDirHandle dir);
void esBindThread2Die(int32_t dieId);
#ifdef __cplusplus
}
#endif

#endif  // ESWIN_HOST_RUNTIME_INCLUDE_UTILS_H_
