// Copyright © 2024 ESWIN. All rights reserved.
//
// Beijing ESWIN Computing Technology Co., Ltd and its affiliated companies ("ESWIN") retain
// all intellectual property and proprietary rights in and to this software. Except as expressly
// authorized by ESWIN, no part of the software may be released, copied, distributed, reproduced,
// modified, adapted, translated, or created derivative work of, in whole or in part.

#ifndef _POSTPROCESS_H
#define _POSTPROCESS_H

#include <stdint.h>
#include <string>
#include <vector>
#include <unordered_map>
#include <mutex>
#include "es_nn_common.h"
#include "es_npu_types.h"

class DetectOutConfigs
{
 public:
    enum EsDetNetwork {
        ES_YOLOV3_U = 0,
        ES_YOLOV3_D = 1,
        ES_YOLOV4 = 2,
        ES_YOLOV5 = 3,
        ES_YOLOV7 = 4,
        ES_YOLOV8 = 5
    };
    enum EsNmsMethod {
        ES_HARD_NMS = 0,
        ES_SOFT_NMS_GAUSSIAN,
        ES_SOFT_NMS_LINEAR
    };

    enum EsIouMethod {
        ES_IOU = 0,
        ES_GIOU,
        ES_DIOU
    };

    enum EsBoxType {
        ES_XMIN_YMIN_XMAX_YMAX = 0,
        ES_XMIN_YMIN_W_H,
        ES_YMIN_XMIN_YMAX_XMAX,
        ES_XMID_YMID_W_H,
    };

    EsDetNetwork detectNet;
    int32_t outTensorNum;
    int32_t inTensorNum;
    std::vector<std::vector<int32_t>> inputShape;
    std::vector<std::vector<int32_t>> outputShape;
    std::vector<int32_t> inputDataType;
    std::vector<int32_t> outputDataType;
    int32_t anchorsNum;
    std::vector<float> anchorScale;
    int32_t imgH;
    int32_t imgW;
    int32_t clsNum;
    std::vector<float> inputScale;
    int32_t nmsMethod;
    int32_t iouMethod;
    int32_t outBoxType;
    int32_t coordNorm;
    int32_t maxBoxesPerClass;
    int32_t maxBoxesPerBatch;
    float scoreThreshold;
    float iouThreshold;
    float softNmsSigma;
    float effecImgOffsetX;
    float effecImgOffsetY;
};

class EsPostProcess
{
 public:
    EsPostProcess();
    virtual ~EsPostProcess();
    bool parseArgs(std::string fileName);
    int32_t getProcessType();
    int32_t getArgmaxK();
    ES_DATA_PRECISION_E convertDataType(uint8_t dataType);
    bool parseClassification(std::string fileName);
    std::string getLabelByIdx(int idx, bool full = false);
    int32_t drawObjectToPicture(std::string &picPath, std::vector<std::vector<float>> &boxinfo,
                                const std::string &outputDir);
    int32_t classify(std::vector<ES_TENSOR_S> inTensors, std::vector<std::vector<int>> &topKIndex,
                     std::vector<std::vector<float>> &topKConfidence);
    int32_t detectionOut(std::vector<ES_TENSOR_S> inTensors, std::vector<std::vector<std::vector<float>>> &boxInfo);
    int32_t segment(std::vector<ES_TENSOR_S> inTensors, void *out);

    typedef enum {
        CLASSIFY = 0,
        DETECTION_OUT,
        SEGMENT,
    } POST_PROCESS_TYPE;

 private:
    int32_t prepareSysMem(ES_U64 &fd, uint64_t size);
    void shapeCopy(ES_TENSOR_S &dst, ES_TENSOR_S &src);
    int32_t convertPrecisionToBytes(ES_DATA_PRECISION_E precision);
    std::vector<std::vector<int32_t>> parseDataShape(std::string &lines);
    std::vector<int32_t> parseDataType(std::string &lines);
    template <typename T>
    std::vector<T> parseArray(std::string &lines);
    template <typename T>
    std::vector<int32_t> extractBoxNum(void *data, int32_t count);
    template <typename T>
    std::vector<std::vector<std::vector<float>>> extractBoxInfo(void *data, std::vector<int32_t> &boxCnt, int32_t size);
    std::unordered_map<int, std::vector<std::string>> mLabelMap;
    std::string mProcessName;
    POST_PROCESS_TYPE mProcessType;
    int32_t mArgmaxK;
    DetectOutConfigs mDetectOutConfigs;
    static int32_t mInstanceCount;
    std::mutex mMtx;
};

#endif  // _POSTPROCESS_H