// Copyright © 2024 ESWIN. All rights reserved.
//
// Beijing ESWIN Computing Technology Co., Ltd and its affiliated companies ("ESWIN") retain
// all intellectual property and proprietary rights in and to this software. Except as expressly
// authorized by ESWIN, no part of the software may be released, copied, distributed, reproduced,
// modified, adapted, translated, or created derivative work of, in whole or in part.

#include <fstream>
#include <chrono>
#include <filesystem>
#include <vector>
#include <algorithm>
#include "hwpreprocess.h"

#define MAX_WIDTH 3840
#define MAX_HEIGHT 2160

std::atomic<bool> EsHwPreProcess::mInit{false};

EsHwPreProcess::EsHwPreProcess() : mGrpCount(1), mGrpId(0), mChnId(0) {}

EsHwPreProcess::~EsHwPreProcess() { release(); }

bool EsHwPreProcess::initializeFromJson(uint32_t devId, const std::string &cfgFilePath)
{
    bool success = EsPreProcess::initializeFromJson(devId, cfgFilePath);
    if (!success) return success;

    // set normalization params
    memset(&mNormParams, 0, sizeof(mNormParams));
    mNormParams.normalizationMode = VPS_NORMALIZATION_Z_SCORE;

    ES_FLOAT mean1, mean2, mean3;
    ES_FLOAT std1, std2, std3;
    ES_FLOAT scale;

    if (mNorm) {
        scale = 255;
    } else {
        scale = 1;
    }

    mean1 = mMeanValues[0] * scale;
    mean2 = mMeanValues[1] * scale;
    mean3 = mMeanValues[2] * scale;
    std1 = 1 / (mStdValues[0] * scale);
    std2 = 1 / (mStdValues[0] * scale);
    std3 = 1 / (mStdValues[0] * scale);
    if (mInputFormat == "BGR") {
        mNormParams.meanValue.r = *((ES_U32 *)&mean3);
        mNormParams.meanValue.g = *((ES_U32 *)&mean2);
        mNormParams.meanValue.b = *((ES_U32 *)&mean1);
        mNormParams.stdReciprocal.r = *((ES_U32 *)&std3);
        mNormParams.stdReciprocal.g = *((ES_U32 *)&std2);
        mNormParams.stdReciprocal.b = *((ES_U32 *)&std1);
    } else {
        mNormParams.meanValue.r = *((ES_U32 *)&mean1);
        mNormParams.meanValue.g = *((ES_U32 *)&mean2);
        mNormParams.meanValue.b = *((ES_U32 *)&mean3);
        mNormParams.stdReciprocal.r = *((ES_U32 *)&std1);
        mNormParams.stdReciprocal.g = *((ES_U32 *)&std2);
        mNormParams.stdReciprocal.b = *((ES_U32 *)&std3);
    }

    ES_FLOAT reciprocal = 1 / mQuantizationScale;
    mNormParams.stepReciprocal = *((ES_U32 *)&reciprocal);
    // mNormParams.bByPassQuantization = ES_TRUE;

    if (!mInit.load()) {
        mInit.store(true);
        ES_S32 i;
        ES_S32 ret = ES_SUCCESS;
        SAMPLE_VDEC_ATTR sampleVdec[ES_VDEC_MAX_GRP_NUM];
        memset(sampleVdec, 0, sizeof(sampleVdec));

        /************************************************
        step1:  init SYS, init vdec
        *************************************************/
        ret = ES_SYS_Init();
        if (ES_SUCCESS != ret) {
            DEBUG_ERROR("ES_SYS_Init failed!\n");
            goto END1;
        }
        ret = SAMPLE_COMM_VDEC_Init();
        if (ES_SUCCESS != ret) {
            DEBUG_ERROR("SAMPLE_COMM_VDEC_Init failed!\n");
            goto END2;
        }

        /************************************************
        step2:  init module VB or user VB(for VDEC)
        *************************************************/
        for (i = 0; i < mGrpCount; i++) {
            sampleVdec[i].type = PT_JPEG;
            sampleVdec[i].width = MAX_WIDTH;
            sampleVdec[i].height = MAX_HEIGHT;
            sampleVdec[i].mode = VIDEO_MODE_FRAME;
            sampleVdec[i].displayFrameNum = 8;
            sampleVdec[i].frameBufCnt = sampleVdec[i].displayFrameNum + 1;
            sampleVdec[i].enableChn[0] = ES_TRUE;
            sampleVdec[i].chnMode[0].pixelFormat = PIXEL_FORMAT_R8G8B8;
            // sampleVdec[i].align = 64;
        }

        ret = SAMPLE_COMM_VDEC_InitVBPool(mGrpCount, &sampleVdec[0]);
        if (ret != ES_SUCCESS) {
            DEBUG_ERROR("init mod common vb fail for %#x!\n", ret);
            goto END3;
        }

        /************************************************
        step3:  start VDEC
        *************************************************/
        ret = SAMPLE_COMM_VDEC_Start(devId, mGrpCount, &sampleVdec[0]);
        if (ret != ES_SUCCESS) {
            DEBUG_ERROR("start VDEC fail for %#x!\n", ret);
            goto END4;
        }

        /************************************************
        step4:  init VPS
        *************************************************/
        ret = SAMPLE_COMM_VPS_Init();
        if (ES_SUCCESS != ret) {
            DEBUG_ERROR("SAMPLE_COMM_VPS_Init failed with %#x!\n", ret);
            goto END5;
        }
    }

    return success;
END5:
    SAMPLE_COMM_VPS_Deinit();

END4:
    SAMPLE_COMM_VDEC_Stop(mGrpCount);

END3:
    SAMPLE_COMM_VDEC_ExitVBPool();

END2:
    SAMPLE_COMM_VDEC_Deinit();

END1:
    ES_SYS_Exit();

    return false;
}

int32_t EsHwPreProcess::preprocess(const std::string &jpegPath, uint64_t fd, uint64_t offset, uint64_t size)
{
    std::lock_guard<std::mutex> lock(mMtx);

    VDEC_STREAM_S stream;
    VIDEO_FRAME_INFO_S frame;
    ES_U64 PTS = 0;
    ES_S32 ret;
    ES_FLOAT scale;
    SIZE_S middSize, dstSize;
    PIXEL_FORMAT_E middFmt, dstFmt;
    RECT_S srcRect, middRect, dstRect;

    auto start_time = std::chrono::high_resolution_clock::now();

    std::ifstream inFileStream(jpegPath.c_str(), std::ios::binary | std::ios::ate);
    if (!inFileStream.is_open()) {
        DEBUG_ERROR("Failed to open file:%s\n", jpegPath.c_str());
        return ES_FAILURE;
    }

    std::streampos inFileLen = inFileStream.tellg();
    inFileStream.seekg(0, std::ios::beg);
    std::vector<char> inFileData(inFileLen);

    if (!inFileStream.read(inFileData.data(), inFileLen)) {
        DEBUG_ERROR("Failed to read file:%s\n", jpegPath.c_str());
        return ES_FAILURE;
    }

    stream.PTS = PTS;
    stream.pAddr = (ES_U8 *)inFileData.data();
    stream.len = inFileLen;
    stream.bEndOfFrame = ES_TRUE;
    stream.bEndOfStream = ES_FALSE;
    stream.bDisplay = ES_TRUE;

    ret = ES_VDEC_SendStream(mGrpId, &stream, -1);
    if (ES_SUCCESS != ret) {
        DEBUG_ERROR("ES_VDEC_SendStream failed\n");
        return ES_FAILURE;
    }

    ret = ES_VDEC_GetFrame(mGrpId, mChnId, &frame, -1);
    if (ES_SUCCESS == ret) {
        // FILE *fp = fopen("dump111.rgb", "w");
        // SAMPLE_COMM_WriteYUV(&frame, fp, 1, 1);
        // fclose(fp);
    } else {
        DEBUG_ERROR("ES_VDEC_GetFrame failed\n");
    }

    VIDEO_FRAME_INFO_S middFrame;
    memset(&middFrame, 0, sizeof(VIDEO_FRAME_INFO_S));
    middSize.width = mResizeShape;
    middSize.height = mResizeShape;
    middFmt = PIXEL_FORMAT_R8G8B8;
    ret = SAMPLE_COMM_INIT_BUFFER(&middFrame, middSize, middFmt);
    if (ret < 0) {
        SAMPLE_PRT("init dstFrame buffer failed, ret: %d\n", ret);
        ret = ES_FAILURE;
        return ret;
    }

    ES_S32 newHeight;
    ES_S32 newWidth;

    if (mKeepRatio) {
        // fill whole image with fillColor
        RECT_S fillRect = {0, 0, middFrame.videoFrame.width, middFrame.videoFrame.height};
        ES_U32 fillColor = 0xFF727272;
        ret = ES_VPS_Fill(&middFrame.videoFrame, &fillColor, 1, &fillRect, 1, HW_TYPE_HAE);
        if (ret != ES_SUCCESS) {
            DEBUG_ERROR("ES_VPS_Fill failed, %d\n", ret);
        }
        scale = static_cast<ES_FLOAT>(mResizeShape) / std::max(frame.videoFrame.width, frame.videoFrame.height);
        newHeight = static_cast<ES_S32>(frame.videoFrame.height * scale);
        newWidth = static_cast<ES_S32>(frame.videoFrame.width * scale);
    } else {
        newHeight = middFrame.videoFrame.height;
        newWidth = middFrame.videoFrame.width;
    }

    memset(&middRect, 0, sizeof(middRect));
    if (frame.videoFrame.width > frame.videoFrame.height) {
        middRect.x = 0;
        middRect.y = (middSize.height - newHeight) / 2;
        middRect.width = newWidth;
        middRect.height = newHeight;
    } else {
        middRect.x = (middSize.width - newWidth) / 2;
        middRect.y = 0;
        middRect.width = newWidth;
        middRect.height = newHeight;
    }

    VPS_PROPERTY_S prop = {};
    prop.type = VPS_PROPERTY_RESIZE_METHOD;
    // VPS_RESIZE_STRETCHBLIT, VPS_RESIZE_FILTERBLIT, VPS_RESIZE_BILINEAR, VPS_RESIZE_BICUBIC
    prop.u.resizeMethod = VPS_RESIZE_FILTERBLIT;
    ret = ES_VPS_SetProperty(&prop);
    if (ret != ES_SUCCESS) {
        DEBUG_ERROR("Set resize method [%d] failed\n", prop.u.resizeMethod);
    }

    // resize frame(originW, originH) to middle frame
    ret = ES_VPS_CropResize(&frame.videoFrame, &middFrame.videoFrame, ES_NULL, &middRect, 0, HW_TYPE_HAE);
    if (ret != ES_SUCCESS) {
        DEBUG_ERROR("ES_VPS_Resize failed, %d\n", ret);
    }

    // 2D support bgr2grb, not rgb2bgr, so we should adjust mean and std, then set input to bgr.
    if (mInputFormat == "BGR") {
        middFrame.videoFrame.pixelFormat = PIXEL_FORMAT_B8G8R8;
    }
    VIDEO_FRAME_INFO_S dstFrame;
    memset(&dstFrame, 0, sizeof(VIDEO_FRAME_INFO_S));
    dstFrame.poolId = ES_VB_INVALID_POOLID;
    dstFrame.modId = ES_ID_USER;
    dstFrame.videoFrame.width = mCropWidth;
    dstFrame.videoFrame.height = mCropWidth;
    dstFrame.videoFrame.fd = fd;

    if (mChannelFirst) {
        dstFrame.videoFrame.pixelFormat = PIXEL_FORMAT_R8G8B8I_PLANAR;
        dstFrame.videoFrame.stride[0] = mCropWidth;
        dstFrame.videoFrame.stride[1] = mCropWidth;
        dstFrame.videoFrame.stride[2] = mCropWidth;

        dstFrame.videoFrame.offset[0] = offset;
        dstFrame.videoFrame.offset[1] = offset + (mCropWidth * mCropWidth);
        dstFrame.videoFrame.offset[2] = offset + (mCropWidth * mCropWidth * 2);
    } else {
        dstFrame.videoFrame.pixelFormat = PIXEL_FORMAT_R8G8B8I;
        dstFrame.videoFrame.stride[0] = mCropWidth * 3;
        dstFrame.videoFrame.stride[1] = 0;
        dstFrame.videoFrame.stride[2] = 0;

        dstFrame.videoFrame.offset[0] = offset;
        dstFrame.videoFrame.offset[1] = 0;
        dstFrame.videoFrame.offset[2] = 0;
    }

    memset(&srcRect, 0, sizeof(srcRect));
    srcRect.x = (middFrame.videoFrame.width - dstFrame.videoFrame.width) / 2;
    srcRect.y = (middFrame.videoFrame.height - dstFrame.videoFrame.height) / 2;
    srcRect.width = dstFrame.videoFrame.width;
    srcRect.height = dstFrame.videoFrame.height;

    DEBUG_INFO("src width:%d, height:%d\n", frame.videoFrame.width, frame.videoFrame.height);
    DEBUG_INFO("dst width:%d, height:%d\n", dstFrame.videoFrame.width, dstFrame.videoFrame.height);
    DEBUG_INFO("src rect: w:%d, h:%d, x:%d, y:%d\n", srcRect.width, srcRect.height, srcRect.x, srcRect.y);

    // Normalization
    ret = ES_VPS_Normalization(&middFrame.videoFrame, &dstFrame.videoFrame, &mNormParams, &srcRect, NULL);
    if (ret != ES_SUCCESS) {
        DEBUG_ERROR("ES_VPS_Normalization return with 0x%x!\n", ret);
    }

    // Release video frame
    ES_VDEC_ReleaseFrame(mGrpId, mChnId, &frame);
    ////////////////////////////////////////////////////////////////////////////////////////
    SAMPLE_COMM_DEINIT_BUFFER(&middFrame);

    auto end_time = std::chrono::high_resolution_clock::now();
    auto duration = std::chrono::duration_cast<std::chrono::duration<double, std::milli>>(end_time - start_time);

    DEBUG_INFO("EsHwPreProcess::%s Preprocess cost:%.4f\n", __func__, duration.count());

    return ret;
}

void EsHwPreProcess::release()
{
    if (mInit.load()) {
        mInit.store(false);
        DEBUG_INFO("EsHwPreProcess::release\n");
        SAMPLE_COMM_VPS_Deinit();
        SAMPLE_COMM_VDEC_Stop(mGrpCount);
        SAMPLE_COMM_VDEC_ExitVBPool();
        SAMPLE_COMM_VDEC_Deinit();
        ES_SYS_Exit();
    }
}
