// Copyright © 2024 ESWIN. All rights reserved.
//
// Beijing ESWIN Computing Technology Co., Ltd and its affiliated companies ("ESWIN") retain
// all intellectual property and proprietary rights in and to this software. Except as expressly
// authorized by ESWIN, no part of the software may be released, copied, distributed, reproduced,
// modified, adapted, translated, or created derivative work of, in whole or in part.

#include <iostream>
#include <fstream>
#include <cstring>
#include <filesystem>
#include <algorithm>
#include <opencv2/opencv.hpp>
#include "es_npu_interface.h"
#include "es_sys_memory.h"
#include "preprocess.h"
#include "cJSON.h"

#define GRAY_PADDING_VALUE 114
#define IVALID_HANDLE -1

EsPreProcess::EsPreProcess() {}

EsPreProcess::~EsPreProcess() {}

bool EsPreProcess::initializeFromJson(uint32_t devId, const std::string &cfgFilePath)
{
    std::ifstream file(cfgFilePath);
    if (!file.is_open()) {
        DEBUG_ERROR("Failed to open JSON file\n");
        return false;
    }

    std::string jsonString((std::istreambuf_iterator<char>(file)), std::istreambuf_iterator<char>());

    cJSON *root = cJSON_Parse(jsonString.c_str());
    if (!root) {
        DEBUG_ERROR("Failed to parse JSON\n");
        return false;
    }

    cJSON *meanItem = cJSON_GetObjectItem(root, "mean");
    cJSON *stdItem = cJSON_GetObjectItem(root, "std");

    if (!meanItem || !stdItem) {
        DEBUG_ERROR("Failed to get mean and std\n");
        cJSON_Delete(root);
        return false;
    }

    if (cJSON_GetArraySize(meanItem) == 3) {
        mMeanValues[0] = cJSON_GetArrayItem(meanItem, 0)->valuedouble;
        mMeanValues[1] = cJSON_GetArrayItem(meanItem, 1)->valuedouble;
        mMeanValues[2] = cJSON_GetArrayItem(meanItem, 2)->valuedouble;
    } else {
        mMeanValues[0] = 123.675;
        mMeanValues[1] = 116.28;
        mMeanValues[2] = 103.53;
    }

    if (cJSON_GetArraySize(stdItem) == 3) {
        mStdValues[0] = cJSON_GetArrayItem(stdItem, 0)->valuedouble;
        mStdValues[1] = cJSON_GetArrayItem(stdItem, 1)->valuedouble;
        mStdValues[2] = cJSON_GetArrayItem(stdItem, 2)->valuedouble;
    } else {
        mStdValues[0] = 0.01712475;
        mStdValues[1] = 0.017507;
        mStdValues[2] = 0.01742919;
    }

    mInputFormat = cJSON_GetObjectItem(root, "input_format")->valuestring;
    mKeepRatio = cJSON_GetObjectItem(root, "keep_ratio")->valueint;
    mChannelFirst = cJSON_GetObjectItem(root, "channel_first")->valueint;
    mNorm = cJSON_GetObjectItem(root, "norm")->valueint;
    mResizeShape = cJSON_GetArrayItem(cJSON_GetObjectItem(root, "resize_shape"), 0)->valueint;
    mCropWidth = cJSON_GetArrayItem(cJSON_GetObjectItem(root, "crop_shape"), 0)->valueint;
    mCropHeight = cJSON_GetArrayItem(cJSON_GetObjectItem(root, "crop_shape"), 1)->valueint;
    mQuantizationScale = cJSON_GetObjectItem(root, "scale")->valuedouble;

    DEBUG_INFO("===============================================\n");
    DEBUG_INFO("Parsed Preprocess Configuration From JSON:\n");
    DEBUG_INFO("mean: [%f, %f, %f]\n", mMeanValues[0], mMeanValues[1], mMeanValues[2]);
    DEBUG_INFO("std: [%f, %f, %f]\n", mStdValues[0], mStdValues[1], mStdValues[2]);
    DEBUG_INFO("input_format: %s\n", mInputFormat.c_str());
    DEBUG_INFO("keep_ratio: %s\n", mKeepRatio ? "true" : "false");
    DEBUG_INFO("channel_first: %s\n", mChannelFirst ? "true" : "false");
    DEBUG_INFO("norm: %s\n", mNorm ? "true" : "false");
    DEBUG_INFO("resize_shape: %d\n", mResizeShape);
    DEBUG_INFO("crop_shape: [%d, %d]\n", mCropWidth, mCropHeight);
    DEBUG_INFO("quantization_scale: %f\n", mQuantizationScale);
    DEBUG_INFO("===============================================\n");

    cJSON_Delete(root);

    return true;
}

int32_t EsPreProcess::preprocess(const std::string &jpegPath, uint64_t fd, uint64_t offset, uint64_t size)
{
    uint8_t *pData = (uint8_t *)ES_SYS_Mmap(fd, size + offset, SYS_CACHE_MODE_NOCACHE);
    if (!pData) {
        printf("ES_SYS_Mmap failed!\n");
        return ES_FAILURE;
    }

    memset(pData, 0, size);
    preprocessInternal(jpegPath, pData + offset, size);
    ES_SYS_Munmap(pData, size + offset);

    return 0;
}

int32_t EsPreProcess::preprocessInternal(const std::string &jpegPath, uint8_t *inputBuf, int32_t size)
{
    auto start_time = std::chrono::high_resolution_clock::now();

    cv::Mat img = cv::imread(jpegPath);
    if (img.empty()) {
        DEBUG_ERROR("Cannot read the image: %s\n", jpegPath.c_str());
        return ES_FAILURE;
    }

    // Convert to RGB as mean and std is RGB
    cv::cvtColor(img, img, cv::COLOR_BGR2RGB);

    // Get original dimensions
    int originalHeight = img.rows;
    int originalWidth = img.cols;

    if (mKeepRatio) {
        // Calculate scaling factor
        double scale = static_cast<double>(mResizeShape) / std::max(originalHeight, originalWidth);

        // Calculate new dimensions
        int newHeight = static_cast<int>(originalHeight * scale);
        int newWidth = static_cast<int>(originalWidth * scale);

        // Resize the image
        cv::Mat imgResized;
        cv::resize(img, imgResized, cv::Size(newWidth, newHeight));

        // Create padded background
        cv::Mat paddedImg = cv::Mat::zeros(mResizeShape, mResizeShape, imgResized.type()) +
                            cv::Scalar(GRAY_PADDING_VALUE, GRAY_PADDING_VALUE, GRAY_PADDING_VALUE);

        // Calculate padding position
        int startX = (mResizeShape - newWidth) / 2;
        int startY = (mResizeShape - newHeight) / 2;

        // Paste the resized image onto the padded background
        imgResized.copyTo(paddedImg(cv::Rect(startX, startY, newWidth, newHeight)));
        img = paddedImg;
    } else {
        cv::resize(img, img, cv::Size(mResizeShape, mResizeShape));
    }

    // Center crop
    int cropStartX = (mResizeShape - mCropWidth) / 2;
    int cropStartY = (mResizeShape - mCropHeight) / 2;
    cv::Rect cropRect(cropStartX, cropStartY, mCropWidth, mCropHeight);
    cv::Mat imgCropped = img(cropRect);

    // Convert to float type and normalize
    imgCropped.convertTo(imgCropped, CV_32F, 1.0 / 255.0);

    cv::Scalar meanValues = cv::Scalar(mMeanValues[0], mMeanValues[1], mMeanValues[2]);
    cv::Scalar stdValues = cv::Scalar(mStdValues[0], mStdValues[1], mStdValues[2]);
    cv::subtract(imgCropped, meanValues, imgCropped);
    cv::divide(imgCropped, stdValues, imgCropped);

    /////////////////////////////////////////////////////////////////////////////////////////////

    if (mInputFormat == "BGR") {
        cv::cvtColor(imgCropped, imgCropped, cv::COLOR_RGB2BGR);
    }

#if DEBUG_MODE
    // Write preprocess data to jpeg
    cv::Mat preprocessedImgToSave;
    imgCropped.convertTo(preprocessedImgToSave, CV_8U, 255.0);
    imwrite("preprocess.jpeg", preprocessedImgToSave);
#endif

    /////////////////////////////////////////////////////////////////////////////////////////////

    // Convert to quantized int8
    cv::Mat quantized_img;
    imgCropped.convertTo(quantized_img, CV_32S, 1.0 / mQuantizationScale);
    quantized_img = min(max(quantized_img, -128), 127);

    cv::Mat int8Img;
    quantized_img.convertTo(int8Img, CV_8S);
    if (mChannelFirst) {
        // If c first, convert to NCHW format, opencv default format is NHWC
        std::vector<cv::Mat> channels(3);
        cv::split(int8Img, channels);
        cv::Mat nchwImg;
        cv::vconcat(channels, nchwImg);
        nchwImg = nchwImg.reshape(1, std::vector<int>{1, 3, mCropWidth, mCropHeight});
        int8Img = nchwImg;
    }

    int32_t outDataLen = int8Img.total() * int8Img.elemSize();
    DEBUG_INFO("preprocess buffer size:%d, inTensor bufferSize:%d\n", outDataLen, size);
    if (outDataLen > size) {
        DEBUG_ERROR("invalid out buffer size\n");
        return ES_FAILURE;
    }
#if DEBUG_MODE
    std::filesystem::path filePath(jpegPath);
    std::string fileName = filePath.filename().string();
    std::string saveFile = fileName + ".bin";
    std::ofstream outFileStream(saveFile, std::ios::out | std::ios::binary);
    outFileStream.write((char *)int8Img.data, outDataLen);
    outFileStream.close();
#endif

    memcpy(inputBuf, int8Img.data, outDataLen);

    auto end_time = std::chrono::high_resolution_clock::now();
    auto duration = std::chrono::duration_cast<std::chrono::duration<double, std::milli>>(end_time - start_time);

    DEBUG_INFO("EsPreProcess::%s Preprocess cost:%.4f\n", __func__, duration.count());

    return ES_SUCCESS;
}

void EsPreProcess::release() {}